
/*
 *  cycling 23 bits
 *   134.4  Hz bitrate
 *  1968.75 Hz samplerate
 *
 *  337 samples per message cycle gives about right speed.
 *  15, sometimes 14, samples per bit.
 *
 *  XXX make one bit deeper, less dc component
 *  XXX add one bit (illegal 23+1), no dc component
 */

#include "head.h"
#include "cfg.h"

extern byte
	dcs_wave[];

/* sine from -90 to +90 degrees, 14 samples */

static const int template[] = {
	-127, -123, -112, -95, -72, -45, -15,
	15, 45, 72, 95, 112, 123, 127
};

static byte slope[14];

#define STEP(n) (word) ((n) * 0x8000UL * 23 / 337)

void
dcs(int code)
{
	dword bits  = dcs_frame(code);
	byte center = RFC_get();
	int gain    = cfg.ctcss_gain; /* XXX, but generation is identical */

	int i;
	word x;
	byte y, *p = dcs_wave;

	for (i = 0; i < 14; ++i)
		slope[i] = (template[i] * gain >> 8) + center;

	/* copy lsbit to left of msbit, for last slope */

	bits |= (bits << 23);

	if (bits & 1)
		y = slope[0];
	else
		y = slope[14-1];

	/* fudge start angle, to get 337 samples instead of 338 */

	x = 0x108;

	do {
		switch ((byte) bits & 3) {
		case 1:
			for (i = 0; i < 14; ++i)
				*p++ = slope[i];

			y = slope[14-1];
			x += STEP(14);
			break;

		case 2:
			for (i = 14; --i >= 0; )
				*p++ = slope[i];

			y = slope[0];
			x += STEP(14);
			break;
		}
		while ((x & 0x8000) == 0) {
			x += STEP(1);
			*p++ = y;
		}
		x &= ~0x8000;
		bits >>= 1;
	} while (p != dcs_wave + 337); /* crash unless the fudge is correct */

	/* XXX start position might need adjusting,
	 * the receiver should see a valid cycle in minimum time */

	dcs_go();
}

