
; sine contains values 1...127,
; low + high tone is 2 ... 254
;
; pwm rate is 4032 kHz per 256 or 15750 Hz
; P8N loop must take 256 T,
; P8E loop must take 512 T
;
; timer mode is programmed to
; interrupt-on-terminal count, lsbyte only.
; tx (and sidetone) audio paths are open.
;
; P8E times include the 1 (or 2 if prefixed) waitstate

	.module dtmf

	.include "iomap.inc"

	.area _BSS

phacc_1: .ds 2
phacc_2: .ds 2

	.area _CODE

_dtmf_timer_mt::             ; (word tone_1, word tone_2)
	pop hl
	pop bc
	pop de
	push de
	push bc                  ; phinc_2
	push hl

	ld hl, # 0
	ld (phacc_1), hl
	ld (phacc_2), hl

	call _disable            ; from now on.

	ld a, # TMR_SEL_MT | TMR_LSB | TMR_INTTC
	out (TMR_CTRL), a
top:
	ld hl, (phacc_2)         ; 16/17 high tone
	add hl, de               ; 11/12 acc += inc
	ld (phacc_2), hl         ; 16/17
	ld l, h                  ;  4/5
	ld h, #>_dtmf_sine       ;  7/8  sine msbyte, page aligned
	ld a, (hl)               ;  7/8  sine[]

	; add a, a ; 4or5 for extreme "twist" XXX needs smaller "center"

	ld hl, (phacc_1)         ; 16/17 low tone
	add hl, bc               ; 11/12 acc += inc
	ld (phacc_1), hl         ; 16/17
	ld l, h                  ;  4/5
	ld h, #>_dtmf_sine       ;  7/8  sine msbyte, page aligned
	add a, (hl)              ;  7/8  sine[]

	out (TMR_MT), a          ; 11/12 low + high tone
	out (WDOG), a            ; 11/12

	ld hl, (phacc_1)         ; 16/17 dummies
	ld hl, (phacc_1)         ; 16/17
	or a                     ;  4/5  to make P8N loop 256 T

	ld a, (_logic)           ; 13/14
	cp # 0                   ;  7/8
	call z, burn_more        ; 10/18 extra needed for P8E

	ld a, # WR0_RESET_ESCINT ;  7/8
	out (SIO_A_CSR), a       ; 11/12 unlatch, fresh values
	in a, (SIO_A_CSR)        ; 11/12
	and # SA_DA              ;  7/8  key still pressed ?
	jp z, top                ; 10/11 pin inverted by SIO

	; stop timer

	ld a, # TMR_SEL_MT | TMR_LSB | TMR_INTTC
	out (TMR_CTRL), a

	call _sio_a_esc          ; pick up changes XXX
	jp _restore              ; RET


burn_more:                   ;     to make P8E loop 512 T
	ex (sp), ix              ; /25
	ex (sp), ix              ; /25

	ex (sp), ix              ; /25
	ex (sp), ix              ; /25

	ex (sp), ix              ; /25
	ex (sp), ix              ; /25

	ex (sp), ix              ; /25
	ex (sp), ix              ; /25

	jr 1$                    ; /13
1$:
	ret                      ; /11


; XXX can do better XXX
;
; _identical_ to above,
; except SA_INT and jp nz


_dtmf_timer_mt_b::           ; (word tone_1, word tone_2)
	pop hl
	pop bc
	pop de
	push de
	push bc                  ; phinc_2
	push hl

	ld hl, # 0
	ld (phacc_1), hl
	ld (phacc_2), hl

	call _disable            ; from now on.

	ld a, # TMR_SEL_MT | TMR_LSB | TMR_INTTC
	out (TMR_CTRL), a
pot:
	ld hl, (phacc_2)         ; 16/17 high tone
	add hl, de               ; 11/12 acc += inc
	ld (phacc_2), hl         ; 16/17
	ld l, h                  ;  4/5
	ld h, #>_dtmf_sine       ;  7/8  sine msbyte, page aligned
	ld a, (hl)               ;  7/8  sine[]

	; add a, a ; 4or5 for extreme "twist" XXX needs smaller "center"

	ld hl, (phacc_1)         ; 16/17 low tone
	add hl, bc               ; 11/12 acc += inc
	ld (phacc_1), hl         ; 16/17
	ld l, h                  ;  4/5
	ld h, #>_dtmf_sine       ;  7/8  sine msbyte, page aligned
	add a, (hl)              ;  7/8  sine[]

	out (TMR_MT), a          ; 11/12 low + high tone
	out (WDOG), a            ; 11/12

	ld hl, (phacc_1)         ; 16/17 dummies
	ld hl, (phacc_1)         ; 16/17
	or a                     ;  4/5  to make P8N loop 256 T

	ld a, (_logic)           ; 13/14
	cp # 0                   ;  7/8
	call z, burn_more        ; 10/18 extra needed for P8E

	ld a, # WR0_RESET_ESCINT ;  7/8
	out (SIO_A_CSR), a       ; 11/12 unlatch, fresh values
	in a, (SIO_A_CSR)        ; 11/12
	and # SA_INT             ;  7/8  changes in keypad ?
	jp nz, pot               ; 10/11 pin inverted by SIO

	; stop timer

	ld a, # TMR_SEL_MT | TMR_LSB | TMR_INTTC
	out (TMR_CTRL), a

	call _sio_a_esc          ; pick up changes XXX
	jp _restore              ; RET

