
/* iomap.inc is a subset of this file */

#ifdef SDCC
#define ior(ioa)       (ioa)
#define iow(ioa, byte) (ioa) = (byte)

#define IO_ADDR(lab, io) __sfr __at io lab;
#else
#define IO_ADDR(lab, io) enum { lab = io; };
#endif

IO_ADDR(PIO_A_DATA, 0x00)
IO_ADDR(PIO_B_DATA, 0x01)
IO_ADDR(PIO_A_CTRL, 0x02)
IO_ADDR(PIO_B_CTRL, 0x03)
IO_ADDR(SIO_A_DATA, 0x10)
IO_ADDR(SIO_B_DATA, 0x11)
IO_ADDR(SIO_A_CSR,  0x12)
IO_ADDR(SIO_B_CSR,  0x13)
IO_ADDR(TMR_LPF,    0x20)
IO_ADDR(TMR_MT,     0x21)
IO_ADDR(TMR_2,      0x22) /* unused */
IO_ADDR(TMR_CTRL,   0x23)
IO_ADDR(DA_RFC,     0x30)
IO_ADDR(DA_TPC,     0x40)
IO_ADDR(AD_RSSI,    0x50)
IO_ADDR(AD_SQL,     0x51)
IO_ADDR(AD_BATT,    0x52)
IO_ADDR(AD_TPC,     0x53)
IO_ADDR(AD_FPM,     0x54)
IO_ADDR(AD_RPM,     0x55)
IO_ADDR(AD_TP4,     0x56)
IO_ADDR(AD_IN7,     0x57)
IO_ADDR(OUT_0,      0x60)
IO_ADDR(OUT_1,      0x70)
IO_ADDR(OUT_2,      0x80)
IO_ADDR(WDOG,       0x90)
IO_ADDR(FX429_DATA, 0xA2)
IO_ADDR(FX429_CSR,  0xA3)
IO_ADDR(CSMEM,      0xB0)
IO_ADDR(CSPROM,     0xC0)
IO_ADDR(CSKEY,      0xD0)

IO_ADDR(MT8888_WD,  0xC0) /* write data,    RS0=0, A1=0 */
IO_ADDR(MT8888_WC,  0xC1) /* write control, RS0=1, A1=0 */
IO_ADDR(MT8888_RD,  0xC2) /* read data,     RS0=0, A1=1 */
IO_ADDR(MT8888_RS,  0xC3) /* read status,   RS0=1, A1=1 */

extern byte
	out_0,
	out_1,
	out_2;

#define out_0_set(what)    out_0_change((what), (what))
#define out_0_clr(what)    out_0_change((what), 0)

#define out_1_set(what)    out_1_change((what), (what))
#define out_1_clr(what)    out_1_change((what), 0)

#define out_2_set(what)    out_2_change((what), (what))
#define out_2_clr(what)    out_2_change((what), 0)

#define pio_a_set(what)    pio_a_change((what), (what))
#define pio_a_clr(what)    pio_a_change((what), 0)

#define pio_a_input(what)  pio_a_change_ddr((what), (what))
#define pio_a_output(what) pio_a_change_ddr((what), 0)

#define pio_b_set(what)    pio_b_change((what), (what))
#define pio_b_clr(what)    pio_b_change((what), 0)

#define pio_b_input(what)  pio_b_change_ddr((what), (what))
#define pio_b_output(what) pio_b_change_ddr((what), 0)

extern void
	out_0_change(byte clear, byte set),
	out_1_change(byte clear, byte set),
	out_2_change(byte clear, byte set),
	out_2_toggle(byte bits),

	pio_a_change(byte clear, byte set),
	pio_b_change(byte clear, byte set),
	pio_a_change_ddr(byte clear, byte set),
	pio_b_change_ddr(byte clear, byte set),

	sio_a_change(byte clear, byte set),
	sio_b_change(byte clear, byte set),
	sio_b_toggle(byte bits),

	timer_mt(word hz),
	timer_2(word hz);

#define O0_VOLUME 0x07 /* speaker volume */
#define O0_INH    0x08 /* speaker off */
#define O0_AUDIOC 0x10
#define O0_CCIRC  0x20
#define O0_MTC    0x40
#define O0_MICM   0x80

#define O1_SRE    0x01
#define O1_SCE    0x02
#define O1_STE    0x04
#define O1_CLK    0x08
#define O1_SD     0x10
#define O1_RAS    0x20
#define O1_TPS    0x40

#ifdef BASTARD
#define O1_TXOFF  0x00
#define O1_TXON   0x80
#else
#define O1_TXOFF  0x80 /* other is dummy */
#define O1_TXON   0x00
#endif

#define O2_ROMA14 0x01 /* select which ROM bank is at 0x8000...0xC000 */
#define O2_ROMA15 0x02
#define O2_ROM0   0x04 /* 2 banks from ROM0, if '1' */
#define O2_SMEM   0x08 /* ROMA16 on P8E (CSMEM elsewhere) */
#define O2_CS1    0x10
#define O2_CS2    0x20
#define O2_CLK    0x40 /* I2C SCL for CU58AF */
#define O2_DP     0x80

#define SA_KKINT  RR0_DCD
#define SA_DA     RR0_CTS	/* DA or I2C /INT */

#define SB_PTT    RR0_CTS
#define SB_LOCAL  RR0_SYNC
#define SB_DMIDLE RR0_DCD	/* 1 means MBUS idle */

#define SB_MON    WR5_DTR   /* Master ON pulse, a/c coupled */


#define PA_CLK2    0x01  /* 1968.75 Hz square wave      */
#define PA_OFFHOOK 0x02  /* state of HOOK pin, 0 onhook */
#define PA_WDR     0x04  /* watchdog was activated      */
#define PA_PWR     0x08  /* on/off button, 1 means off  */
#define PA_CCIR    0xF0  /* FX003 data                  */

#define PB_RAMA12 0x01	/* S#/L ? P8E: no effect   out               */
#define PB_EXIN1  0x02	/* aka /POR                out 0 if SERV     */
#define PB_EXIN2  0x04  /* aka /IGN aka /EMG       out 0 if GPio2b 0 */
#define PB_DCU    0x08	/* I2C SDA for CU58AF      out 0 if SDA 0    */
#define PB_RXON   0x10  /* /RXON ? wassit ?        out               */
#define PB_TMR0   0x20  /* from timer 2 output     in (old CTCSS in) */
#define PB_EXAL   0x40	/*                         out               */
#define PB_PWROFF 0x80	/*                         out               */

#define FX429_TXENB  0x01
#define FX429_TXPAR  0x02
#define FX429_RXENB  0x04
#define FX429_RXFMT  0x08
#define FX429_TIMER  0xF0

#define FX429_RXRDY  0x01
#define FX429_RXTRUE 0x02
#define FX429_DCD    0x04
#define FX429_TXRDY  0x08
#define FX429_TXIDL  0x10
#define FX429_TMRINT 0x20
#define FX429_SYNC   0x40
#define FX429_SYNT   0x80

/* all gates are +5v */

#define TMR_SEL_LPF 0x00
#define TMR_SEL_MT  0x40
#define TMR_SEL_2   0x80
#define TMR_INTTC   0x00	/* mode 0, intr on terminal cnt */
#define TMR_ONESHOT 0x02	/* mode 1, one shot */
#define TMR_RATEGEN 0x04	/* mode 2, rate generator */
#define TMR_SQWAVE  0x06	/* mode 3, square wave */
#define TMR_SWTRIGG 0x08	/* mode 4, s/w triggered strobe */
#define TMR_HWTRIGG 0x0A	/* mode 5, h/w triggered strobe */
#define TMR_LATCH   0x00
#define TMR_LSB     0x10	/* r/w counter LSB */
#define TMR_MSB     0x20	/* r/w counter MSB */
#define TMR_BOTH    0x30	/* r/w counter 16 bits, LSB first */

#define WR0_SEND_ABORT			0x08
#define WR0_RESET_ESCINT		0x10
#define WR0_CHANNEL_RESET		0x18
#define WR0_REARM_1STC			0x20
#define WR0_RESET_TXINT			0x28
#define WR0_ERROR_RESET			0x30
#define WR0_RETINT				0x38
#define WR0_RESET_RXCRC			0x40
#define WR0_RESET_TXCRC			0x80
#define WR0_RESET_UNDERRUN		0xc0

#define WR1_ESCINT_ENB			0x01
#define WR1_TXINT_ENB			0x02
#define WR1_STATUS_VECTOR		0x04
#define WR1_RXINT_DISABLE		0x00
#define WR1_INT_1STC			0x08
#define WR1_INT_ALLC			0x10
#define WR1_INT_ALLC_PNS		0x18
#define WR1_WAITREADY_RT		0x20
#define WR1_WAITREADY_FN		0x40
#define WR1_WAITREADY_ENB		0x80

#define WR3_RX_ENB				0x01
#define WR3_SYNCLOAD_INHIBIT	0x02
#define WR3_ADDRESS_SEARCH_MODE	0x04
#define WR3_RXCRC_ENB			0x08
#define WR3_ENTER_HUNT			0x10
#define WR3_AUTO_ENABLES		0x20
#define WR3_5BIT_RX				0x00
#define WR3_6BIT_RX				0x40
#define WR3_7BIT_RX				0x80
#define WR3_8BIT_RX				0xc0

#define WR4_PARITY_ENB			0x01
#define WR4_PARITY_ODD			0x00
#define WR4_PARITY_EVEN			0x02
#define WR4_SYNC_MODES			0x00
#define WR4_1STOPBIT			0x04
#define WR4_15STOPBIT			0x08
#define WR4_2STOPBIT			0x0c
#define WR4_8BITSYNC			0x00
#define WR4_16BITSYNC			0x10
#define WR4_SDLC_MODE			0x20
#define WR4_EXTSYNC				0x30
#define WR4_X1_CLK				0x00	/* 153600 bd, but... */
#define WR4_X16_CLK				0x40	/*   9600 bd */
#define WR4_X32_CLK				0x80	/*   4800 bd */
#define WR4_X64_CLK				0xc0	/*   2400 bd */

#define WR5_TXCRC_ENB			0x01
#define WR5_RTS					0x02
#define WR5_CCITT_CRC			0x00
#define WR5_CRC16_CRC			0x04
#define WR5_TX_ENB				0x08
#define WR5_SEND_BREAK			0x10
#define WR5_5BIT_TX				0x00
#define WR5_7BIT_TX				0x20
#define WR5_6BIT_TX				0x40
#define WR5_8BIT_TX				0x60
#define WR5_DTR					0x80

#define RR0_RCA					0x01
#define RR0_INT_PENDING			0x02
#define RR0_TBE					0x04
#define RR0_DCD					0x08
#define RR0_SYNC				0x10
#define RR0_CTS					0x20
#define RR0_UNDERRUN			0x40
#define RR0_BREAK_ABORT			0x80

#define RR1_ALL_SENT			0x01

#define RR1_PARER				0x10
#define RR1_ROVER				0x20
#define RR1_FRMERR				0x40
#define RR1_CRCERR				0x40
#define RR1_ENDFRAME			0x80

