
;
; return value in L, HL or DEHL
; all regs except ix (iy ?) can be trashed
; args in stack (byte promotion ?)
;
; zero count means 0x10000 in many cases
;
	.module util

	.area _CODE

;----------------------------------------------------------------------

_abs::             ; word abs(int);
	pop de
	pop hl
	push hl
	push de

	ld a, h
	or a
	jp p, 1$       ; already positive
	cpl
	ld h, a
	ld a, l
	cpl
	ld l, a
	inc hl
1$:
	ret

_labs::            ; dword labs(long);
	pop bc
	pop hl
	pop de
	push de
	push hl
	push bc

	ld a, d
	or a
	jp p, 1$       ; already positive
	cpl
	ld d, a
	ld a, e
	cpl
	ld e, a
	ld a, h
	cpl
	ld h, a
	ld a, l
	cpl
	ld l, a
	inc l
	jr nz, 1$
	inc h
	jr nz, 1$
	inc de
1$:
	ret

;----------------------------------------------------------------------

_strcmp::          ; short strcmp(byte *, byte *);
	pop bc
	pop de
	pop hl
	push hl
	push de
	push bc
1$:
	ld a, (de)
	inc de
	cpi            ; compared, de and hl incremented
	jr nz, 2$      ; difference
	or a
	jp nz, 1$      ; while same, but not \0

	ld hl, # 0     ; same.
	ret
2$:
	dec hl
	sub (hl)
	ld l, a        ; LSByte of result
	sbc a, a       ; 0 or FF
	ld h, a
	ret

;----------------------------------------------------------------------

_index::           ; byte *index(byte *, word c);
	pop de
	pop hl
	pop bc         ; byte in c
	push bc
	push hl
	push de
1$:
	ld a, (hl)
	cp c
	ret z          ; HL points to the byte, can be \0
	inc hl
	or a
	jp nz, 1$

	ld hl, # 0     ; NULL
	ret

;----------------------------------------------------------------------

_rindex::          ; byte *rindex(byte *, word c);
	pop hl
	pop de
	pop bc         ; byte in c, \0 (silly) ok
	push bc
	push de
	push hl

	ld hl, # 0     ; NULL
1$:
	ld a, (de)
	cp c
	jp nz, 2$
	ld l, e
	ld h, d        ; could be this
2$:
	inc de
	or a
	jp nz, 1$

	ret            ; NULL or not, in HL

;----------------------------------------------------------------------

_memchr::          ; byte *memchr(void *, word c, word len);
	pop af
	pop hl
	pop de         ; byte in e
	pop bc
	push bc
	push de
	push hl
	push af

	ld a, e
	cpir
	dec hl         ; back to the byte
	ret z          ; if found

	ld hl, # 0     ; else NULL
	ret

;----------------------------------------------------------------------

_strend::          ; byte *strend(byte *);
	pop de
	pop hl         ; string
	push hl
	push de

	ld bc, # 0
	xor a
	cpir
	dec hl         ; back over \0
	ret            ; end in HL

;----------------------------------------------------------------------

_strlen::          ; word strlen(byte *);
	pop de
	pop hl         ; string
	push hl
	push de

	ld bc, # 0
	xor a
	cpir
	ld hl, # -1
	sbc hl, bc     ; carry still clear from xor
	ret            ; length in HL

;----------------------------------------------------------------------

_strcpyx::         ; byte *strcpyx(byte *dst, byte *src);
	pop bc
	pop de
	pop hl         ; src
	push hl
	push de
	push bc

	xor a
1$:
	cp (hl)
	ldi            ; keeps Z
	jp nz, 1$

	dec de         ; back over \0
	ex de, hl
	ret            ; end of dst in HL

;----------------------------------------------------------------------

_strncpyx::        ; byte *strncpyx(byte *dst, byte *src, word sz);
	pop af
	pop de
	pop hl
	pop bc
	push bc
	push hl
	push de
	push af

	dec bc         ; leave room for \0
	xor a
1$:
	cp (hl)        ; \0 ?
	jr z, 2$
	ldi
	jp pe, 1$      ; while sz
2$:
	ld (de), a
	ex de, hl
	ret            ; end of dst in HL

;----------------------------------------------------------------------

_memset::          ; void memset(void *, word c, word len);
	pop af
	pop hl
	pop de         ; word c
	pop bc         ; len
	push bc
	push de
	push hl
	push af

	ld (hl), e
	dec bc
	ld a, c
	or b
	ret z
	ld e, l
	ld d, h
	inc de
	ldir
	ret

;----------------------------------------------------------------------

_bzero::           ; void bzero(void *, word len);
	pop hl
	pop de
	pop bc         ; len
	push bc
	push de
	push hl        ; ra

	ld a, c
	or b
	ret z          ; len 0

	ld l, e
	ld h, d
	ld (hl), # 0
	dec bc
	ld a, c
	or b
	ret z          ; len 1
	inc de
	ldir
	ret

;----------------------------------------------------------------------

_memcpy::          ; void memcpy(void *dst, void *src, word len);
	pop af
	pop de
	pop hl
	pop bc
	push bc
	push hl
	push de
	push af
	ldir
	ret

_bcopy::           ; void bcopy(void *src, void *dst, word len);
	pop af
	pop hl
	pop de
	pop bc
	push bc
	push de
	push hl
	push af

	ld a, c
	or b
	ret z          ; len 0

	and a
	sbc hl, de     ; carry if src < dst. not bothering with RET Z
	add hl, de
	jp c, 1$       ; copy backwards
	ldir
	ret
1$:
	add hl, bc
	ex de, hl
	add hl, bc
	ex de, hl
	dec hl
	dec de
	lddr
	ret

;----------------------------------------------------------------------

_bcmp::            ; byte/bool bcmp(void *, void *, word len);
	pop af
	pop hl
	pop de
	pop bc
	push bc
	push de
	push hl
	push af

	ld a, c
	or b
	jr z, 2$       ; len 0, same
1$:
	ld a, (de)
	inc de
	cpi
	jr nz, 3$      ; different
	jp pe, 1$      ; while len
2$:
	ld l, # 0
	ret            ; same.
3$:
	ld l, # 1
	ret            ; different

;----------------------------------------------------------------------

_crc_ccitt::       ; word crc_ccitt(byte *, word);
	pop hl
	pop de
	pop bc
	push bc
	push de
	push hl

	ld b, c        ; length is 1...256 XXX

	ld a, # 0xFF
	ld c, a        ; iv FFFF

	ld h, #>ccitt_tab_hi_lo
1$:
	ex de, hl
	xor (hl)
	inc hl
	ex de, hl

	ld l, a
	ld a, c
	ld c, (hl)
	inc h
	xor (hl)
	dec h
	djnz 1$

	cpl
	ld l, a
	ld a, c
	cpl
	ld h, a        ; ~crc in HL
	ret

;----------------------------------------------------------------------

; HLBC / DE, 31 bits, 15 bits
; HL remainder
; BC quotient

div_bc_de:
	ld hl, # 0

div_hlbc_de:

	ld a, b           ; HLAC
	ld b, # 16        ; 16 times
2$:
	rl c              ; HLAC <<= 1, 1st carry flows out from result
	rla
	adc hl, hl        ; original bit 31 must have been 0
	sbc hl, de
	jr nc, 1$         ; ok
	add hl, de        ; carry always reappears
1$:
	djnz 2$           ; 16 times

	rl c              ; the final bit of quotient
	rla

	cpl
	ld b, a
	ld a, c
	cpl
	ld c, a
	ret

;----------------------------------------------------------------------

_ulmoddiv::
	pop de
	pop iy
	pop bc
	push bc
	push iy
	push de

	ld d, 3(iy)
	ld e, 2(iy)
	ld h, 1(iy)
	ld l, 0(iy)

	ld a, c
	cp # 128         ; typical prescalers XXX not worth much
	jr z, md_128
	cp # 64
	jr z, md_64
	neg
	ld c, a          ; produce inverted carries
	xor a
	ld b, # 32
2$:
	adc hl, hl       ; 15
	rl e             ;  8
	rl d             ;  8
	rla              ;  4
	add c
	jr c, 1$
	sub c
1$:
	djnz 2$

	adc hl, hl
	rl e
	rl d

	ld 3(iy), d
	ld 2(iy), e
	ld 1(iy), h
	ld 0(iy), l

	ld l, a
	ret

md_128:
	add hl, hl
	rl e
	rl d
	rla            ; was 128, topbit disappears

	ld 3(iy), a
	ld 2(iy), d
	ld 1(iy), e
	ld 0(iy), h

	srl l          ; return input & 127
	ret

md_64:
	add hl, hl
	rl e
	rl d
	rla             ; was 64
	add hl, hl
	rl e
	rl d
	rla             ; topbits disappear

	ld 3(iy), a
	ld 2(iy), d
	ld 1(iy), e
	ld 0(iy), h

	srl l
	srl l           ; return input & 63
	ret

_ulmoddiv_10::
	pop de
	pop iy
	push iy
	push de

	ld d, 3(iy)
	ld e, 2(iy)
	ld h, 1(iy)
	ld l, 0(iy)

	ld bc, # 0x20F6   ; 32 and -10
	xor a
2$:
	adc hl, hl
	rl e
	rl d
	rla
	add c
	jr c, 1$
	sub c
1$:
	djnz 2$

	adc hl, hl
	rl e
	rl d

	ld 3(iy), d
	ld 2(iy), e
	ld 1(iy), h
	ld 0(iy), l

	ld l, a
	ret

_umoddiv::
	pop de
	pop iy
	pop bc
	push bc
	push iy
	push de

	ld h, 1(iy)
	ld l, 0(iy)

	ld a, c
	neg
	ld c, a
	xor a
	ld b, # 16
2$:
	adc hl, hl
	rla
	add c
	jr c, 1$
	sub c
1$:
	djnz 2$

	adc hl, hl

	ld 1(iy), h
	ld 0(iy), l

	ld l, a
	ret

_umoddiv_10::
	pop de
	pop iy
	push iy
	push de

	ld h, 1(iy)
	ld l, 0(iy)

	ld bc, # 0x10F6   ; 16 and -10
	xor a
2$:
	adc hl, hl
	rla
	add c
	jr c, 1$
	sub c
1$:
	djnz 2$

	adc hl, hl

	ld 1(iy), h
	ld 0(iy), l

	ld l, a
	ret

;----------------------------------------------------------------------
; while (--arg)
;   ;

_delay_usec::          ; void delay_usec(word usec);
	pop hl
	pop de             ; arg
	push de
	push hl

	srl d              ; half
	rr e
	ld a, (_logic)
	or a
	jr z, 1$           ; P8E, loop takes 2.4 usec
	xor a              ; P8N, loop takes 4.9 usec
	srl d              ; quarter
	rr e
1$:
	dec de
	or d
	jp p, 1$           ; 20 or 23 T total

	ret

;----------------------------------------------------------------------
; while (time - mark < 0)
;   ;

_delay_frac::          ; void delay_frac(word fraction_of_second);
	pop hl
	pop de             ; arg
	push de
	push hl

	ld hl, (_time_frac)
	add hl, de
	ld e, l
	ld d, h            ; end time in de
1$:
	and a
	ld hl, (_time_frac)
	sbc hl, de
	jp m, 1$

	ret

;----------------------------------------------------------------------

_timex_get::           ; dword timex_get();
1$:
	ld hl, (_timex+0)
	ld de, (_timex+2)
	ld a,  (_timex+0)
	cp l
	ret z
	jp 1$              ; changed, try again

_time_get::            ; word time_get();
	ld hl, (_time)
	ret

_timef_get::           ; word timef_get();
	ld hl, (_timef)
	ret

_timef_since::
	pop hl
	pop de
	push de
	push hl

	and a
	ld hl, (_timef)
	sbc hl, de
	ret                ; time elapsed, since arg

;----------------------------------------------------------------------

_timef_mark::          ; word timef_mark(word interval);
	pop hl
	pop de
	push de
	push hl

	ld hl, (_timef)
	add hl, de
	set 0, l           ; always nonzero result
	ret

_time_mark::           ; word time_mark(word interval);
	pop hl
	pop de
	push de
	push hl

	ld hl, (_time)
	add hl, de
	set 0, l           ; always nonzero result
	ret

;----------------------------------------------------------------------

_timef_passed::        ; bool timef_passed(word mark);
	pop hl
	pop de
	push de
	push hl

	ld hl, (_timef)
	and a
	sbc hl, de         ; timef - mark
	jp m, ret_false

	ld l, # 1          ; TRUE
	ret

_time_passed::         ; bool time_passed(word mark);
	pop hl
	pop de
	push de
	push hl

	ld hl, (_time)
	and a
	sbc hl, de         ; time - mark
	jp m, ret_false

	ld l, # 1          ; TRUE
	ret

ret_false:
	ld l, # 0          ; FALSE
	ret

;----------------------------------------------------------------------

_hexdig::
	.ascii "0123456789ABCDEF"

;----------------------------------------------------------------------

_nybble::              ; byte/word nybble(char);
	pop hl
	pop bc
	push bc
	push hl

	ld a, c

	sub # '0
	jr c, 2$
	cp # 10
	jr c, 1$           ; 0...9

	sub # 'A-'0
	and # ~0x20        ; upper or lower
	cp # 6
	jr nc, 2$

	add # 10
1$:
	ld l, a
	ld h, # 0
	ret
2$:
	ld hl, # -1
	ret

;----------------------------------------------------------------------

_store_word::      ; void store_word(word address, word value);
	pop de
	pop bc
	pop hl
	push hl
	push bc
	push de

	ld d, # 0xC9   ; RET
	ld e, b
	ld b, c
	ld c, # 0x22   ; LD (NN), HL
	push de
	push bc
	ld iy, # 0
	add iy, sp
	rst 0x20       ; CALL (IY)

	pop af
	pop af
	ret

;----------------------------------------------------------------------

_setjmp::          ; int setjmp(jmp_buf b);
	pop de         ; ra
	pop iy         ; buf { sp pc ix }
	push iy        ; any junk would do XXX push junk instead
	push de

	ld hl, # 4     ; ra and arg not counted
	add hl, sp
	ld 0(iy), l    ; sp
	ld 1(iy), h
	ld 2(iy), e    ; pc
	ld 3(iy), d
	push ix
	pop hl
	ld 4(iy), l    ; ix
	ld 5(iy), h

	ld hl, # 0
	ret

_longjmp::         ; void longjmp(jmp_buf b, int retval);
	pop de         ; ra, discarded
	pop iy         ; buf { sp pc ix }
	pop de         ; retval

	ld b, 5(iy)
	ld c, 4(iy)    ; ix
	push bc
	ld b, 3(iy)
	ld c, 2(iy)    ; pc
	ld h, 1(iy)
	ld l, 0(iy)    ; sp
	pop ix
	ld sp, hl      ; after all data loaded, buf might vanish
	push iy        ; mimics setjmp args
	push bc        ; pc

	ex de, hl      ; retval to hl
	ret

;----------------------------------------------------------------------

_axaddr::          ; byte *axaddr(byte *dst, byte *asc);
	pop bc
	pop de
	pop hl
	push hl
	push de
	push bc

	ld b, # 6
3$:
	ld a, (hl)
	inc hl
	cp # 'a
	jr nc, 2$      ; a..z
	cp # '0
	jr nc, 1$      ; 0..9 or A..Z
	dec hl         ; stop at - or \0
	xor a          ; put blanks
2$:
	xor # 0x20     ; lower to upper, \0 to blank
1$:
	add a, a       ; 4T sla a
	ld (de), a
	inc de
	djnz 3$        ; 6 characters

	ld a, (hl)     ; "-SSID" ?
	or a
	jr z, 5$       ; no ssid
	cp # '-
	jr nz, 4$      ; accept "aa9bbb15" too
	inc hl
4$:
	push de        ; save dst
	push hl
	call _atob     ; byte atob(byte *)
	ld a, l
	pop de
	pop de
5$:
	rla
	and # 0x1E     ; ssid bits
	or  # 0x60     ; reserved
	ld (de), a
	inc de
	ld l, e
	ld h, d        ; return end
	ret

;----------------------------------------------------------------------

_umulbb::               ;; word umulbb(byte, byte);
	pop hl
	pop de
	push de
	push hl

	ld h, d
umulbb:                 ; HL = H * E
	ld l, # 0
	ld d, l
	ld b, # 8
2$:
	add hl, hl
	jr nc, 1$
	add hl, de
1$:
	djnz 2$

	ret

;----------------------------------------------------------------------
; checksum, split and store fields to _nmea[]

_nmea_split::
	pop de
	pop hl              ; "C" string
	push hl
	push de

	ld iy, # _nmea      ; field array
	ld bc, # 0x0F00     ; b nfld  c chk
	ld de, # 0x2C2A     ; d comma e star
	jp 3$
8$:
	inc b               ; back to 0
1$:
	inc hl
2$:
	xor c
	ld c, a
3$:
	ld a, (hl)
	cp e
	jr z, 4$          ; found the star
	or a
	ret z             ; \0. no star
	cp d
	jp nz, 1$         ; not comma, next

	ld (hl), # 0      ; cut at comma

	dec b
	jp m, 8$          ; array full, next

	inc hl            ; next field is here
	ld 0(iy), l
	ld 1(iy), h
	inc iy
	inc iy
	jp 2$
4$:
	ld (hl), # 0      ; cut at star

	call 6$           ; does not return if bad hex
	rlca
	rlca
	rlca
	rlca
	xor c
	ld c, a

	call 6$           ; does not return if bad hex
	xor c
	ret nz            ; checksum wrong

	ld hl, # 0        ; success return

	dec b
	ret m             ; array already full
5$:
	ld 0(iy), l       ; add NULLs at end
	ld 1(iy), h
	inc iy
	inc iy
	dec b
	jp p, 5$          ; keep adding NULLs
	ret

6$:
	inc hl
	ld a, (hl)
	sub # '0
	jr c, 7$
	cp # 10
	ret c             ; 0...9

	sub # 'A-'0
	cp # 6
	jr nc, 7$
	add # 10          ; A...F
	ret

7$:
	pop af         ; abort one level
	ret            ; hl nz

;----------------------------------------------------------------------
; setup thinks all numbers are base 10,
; but this one is base 8, max 0777.
; HL in/out

fix_octal:
	ld a, l
	sub # 778
	ld a, h
	sbc # 778 >> 8
	jr c, 1$

	ld hl, # 0x01FF
	ret                 ; overflow. return max
1$:
	ld de, # 0xFFFF     ; -1 in D and E
	ld bc, # -100
2$:
	inc d
	add hl, bc
	jp c, 2$
	sbc hl, bc

	ld a, l             ; max 99 now
	ld c, # 10
3$:
	inc e
	sub c
	jp nc, 3$
	add c               ; bit triplets in D E and A

	rrca                ; position lsbits to 0xE0
	rrca
	rrca
	or e                ; lsbits 0 0 middlebits
	rrca
	rrca
	rrca                ; middlebits lsbits 0 0
	srl d
	rra
	srl d
	rra
	ld h, d             ; zeroextended msbit
	ld l, a             ; 8 lsbits
	ret

;----------------------------------------------------------------------

_dcs_frame::            ; dword dcs_frame(int code);
	pop de
	pop hl
	push hl
	push de

	push hl             ; code +- 0...777
	call _abs           ; argument not popped, will be used later
	call fix_octal      ; in HL. code was OCTAL, not a decimal number

	; binary rightjust +code in HL

	; frame looks like P11...1 100 C9...1
	; and C1 is sent first
	;
	; get the full 23 bits to E:H:L
	;
	; E   0  P11 P10 P9 - P8  P7  P6  P5
	; H   P4 P3  P2  P1 - 1   0   0   C9
	; L   C8 C7  C6  C5 - C4  C3  C2  C1

	ld de, # 0x8000   ; d collects P1...4 from left and e P11..5 from right

	; mucho gracias MMI Technical Resource website. good shit.

	; P1 = C1 + C2 + C3 + C4 + C5 + C8

	ld a, l           ; 8 lsbits
	and # 0x9F        ; AND clears carry and tests parity
	jp pe, .+4        ; if even parity, skip SCF
	scf               ; CCF would also do
	rr d              ; insert P1

	; P2 = NOT ( C2 + C3 + C4 + C5 + C6 + C9 )

	ld a, h
	rra
	ld a, l
	rra
	and # 0x9F
	jp po, .+4
	scf
	rr d

 	; P3 = C1 + C2 + C6 + C7 + C8

	ld a, l
	and # 0xE3
	jp pe, .+4
	scf
	rr d

 	; P4 = NOT ( C2 + C3 + C7 + C8 + C9 )

	ld a, h
	rra
	ld a, l
	rra
	and # 0xE3
	jp po, .+4
	scf
	rr d

	; P11 = NOT ( C1 + C2 + C3 + C4 + C7 )

	ld a, l
	and # 0x4F
	jp po, .+4
	scf
	rl e

	; P10 = NOT ( C4 + C6 + C7 + C9 )

	ld a, h
	rra
	ld a, l
	rra
	and # 0xB4
	jp po, .+4
	scf
	rl e

 	; P9 = C3 + C5 + C6 + C8 + C9

	ld a, h
	rra
	ld a, l
	rra
	and # 0xDA
	jp pe, .+4
	scf
	rl e

 	; P8 = C2 + C4 + C5 + C7 + C8 + C9

	ld a, h
	rra
	ld a, l
	rra
	and # 0xED
	jp pe, .+4
	scf
	rl e

 	; P7 = C1 + C3 + C4 + C6 + C7 + C8 + C9

	ld a, l
	and # 0xED     ; C8...C1
	jp pe, .+4
	scf            ; C1 +...+ C8

	bit 0, h
	jr z, .+3
	ccf            ; + C9
	rl e

 	; P6 = NOT ( C1 + C4 + C5 + C6 + C8 )

	ld a, l
	and # 0xB9
	jp po, .+4
	scf
	rl e

 	; P5 = NOT ( C1 + C2 + C5 + C9 )

	ld a, h
	rra
	ld a, l
	rla          ; drop C8, A contains C7 ... C1 C9
	and # 0x27
	jp po, .+4
	scf
	rl e

	; whew.

	ld a, h        ; just C9
	or d           ; add 1 0 0 flag and parity lsbits
	ld h, a

	pop af         ; was argument for abs() long time ago
	rla
	jr nc, 1$      ; +code
	ld a, l
	cpl
	ld l, a
	ld a, h
	cpl
	ld h, a
	ld a, e
	cpl
	ld e, a        ; -code, bits inverted
1$:
	res 7, e
	ld d, # 0
	ret            ; in 23 lsbits of DEHL

;----------------------------------------------------------------------

_make_sine::        ; void make_sine(byte *tab, byte gain, byte center)
	pop hl
	pop bc          ; bc tab (c zero)
	pop de          ; d center and e gain
	push de
	push bc
	push hl

	ld a, d         ; center
	or a            ; less than 128 ?
	jp p, 1$        ; distance to 0
	cpl             ;  to 255 is shorter
1$:
	add a, a        ; twice
	cp e            ; gain
	jr nc, 2$
	ld e, a         ; override gain
2$:
3$:
	ld h, #>_sine
	ld l, c         ; tab is page-aligned
	ld h, (hl)
	push bc
	push de
	call umulbb     ; HL=E*H, gain * sine[i] / 256 to h
	pop de
	pop bc

	ld a, d
	sub h           ; center - value
	set 7, c
	ld (bc), a      ; tab[i+128]

	ld a, d         ; center
	add h           ; center + value
	res 7, c
	ld (bc), a      ; tab[i]

	inc c           ; i++
	jp p, 3$        ; halfcycle, 128 times
	ret

;----------------------------------------------------------------------

