
; don't know if the glitch in P/V flag after LD A, I can happen
; in TMPZ84C00AT-8. doing disable/restore nesting by hand anyway.
;
; Sign Zero - HalfCarry - ParityEven/oVerflow add/sub Carry

	.module boot

	.include "iomap.inc"

	.area _MEMORY (ABS) ; ordering of segments for the linker
	.area _CODE
	.area _GSINIT
	.area _GSFINAL
	.area _DATA
	.area _BSS

;----------------------------------------------------------------------
; above C000 is either
;  - 16 kB work ram
;  -  8 kB bbram, visible in C000 and E000. lower 4kB used

	.area _MEMORY (ABS)

	.org 0x8000     ; banked rom area
_rom_bank::

	.org 0xC000     ; 4 kB reserved for bbram, cfg etc
_ram::
_bbram::

	.org 0xC100
_cfg_START::
_cfg::              ; 2304 bytes

	.org 0xCA00
_cfg_END::
_memories::         ; 1280 bytes

	.org 0xCF00
_rfc_table::        ; 256 bytes

	.org 0xD000     ; rest of 16 kB ram here (_data)

	.org 0xF700
_stack::            ; well above 0xF000 (nmi can spike into bbram)
_afsk_sine::

	.org 0xF800
_dtmf_sine::

	.org 0xF900
_ctcss_sine::

	.org 0xFA00
	.org 0xFC00 - 337      ; 0xF[AB]??, bit 2 of MSByte clear
_dcs_wave::                ; and ends in FC00, bit 2 of MSByte set

	.org 0xFC00
_sio_a_rx_buf:      ; sio buffers [256]

	.org 0xFD00
_sio_a_tx_buf:

	.org 0xFE00
_sio_b_rx_buf::

	.org 0xFF00
_sio_b_rx_bufend::
_sio_b_tx_buf::

	.org 0x0000
_sio_b_tx_bufend::

;----------------------------------------------------------------------

	.org 0          ; RST 0   32 kB non-banked rom area
_rom::
reset:
_boot::
_exit::             ; there is no exit
	ld sp, # _stack ; quickly. always executes before /NMI is sampled ? XXX
	di
	jp continue

	.org 0x08       ; RST 08h NOP (SDCC runtime (?))
_nop::
	ret

	.org 0x10       ; RST 10h JP (HL)
	jp (hl)

	.org 0x18       ; RST 18h JP (IX)
	jp (ix)

	.org 0x20       ; RST 20h JP (IY)
	jp (iy)

	.org 0x28       ; RST 28h RETI
	reti

	.org 0x30       ; RST 30h DISABLE
_disable::
	di
	push hl
	ld hl, # DInest
	inc (hl)
	pop hl
	ret nz          ; else oops, too many nested disables

	.org 0x38       ; RST 38h (opcode 0xFF)
oops:
	jr _boot

JPHL    = 0x10
RETI    = 0x28
DISABLE = 0x30

;----------------------------------------------------------------------

_restore::          ; often a JP _RESTORE, so not a restart
	push hl
	ld hl, # DInest
	dec (hl)
	pop hl
	ret nz
	ei
	ret

__enable::
	push af
	xor a
	ld (DInest), a
	pop af
	ei
	ret

	nop
	.word 0,0,0,0,0,0,0,0,0,0     ; ends at 0x0060
bombstack:

	.org 0x66       ; NMI
	call pusha      ; if bombstack was active, jumps to _boot

	ld a, i         ; sets Z and P/V. IFF1=0 prevents the P/V glitch
	jr z, _boot     ; not booted properly. try again

	ld hl, # nmi_ei_ret
	jp pe, 1$              ; IFF2 is 1
	ld hl, # nmi_ret       ;  is 0
1$:
	; might have been reading "DTMF-prom" (ROM1)
	; or second rom bank (top 16 kB of ROM0);
	; force rom-space back to contiguous 48 kB.
	;
	; difficult to force P8N SMEM=1,
	; as out_2 (and logic) live in work ram.
	; XXX just boot from bombstack for now.
	;
	; XXX maybe put out_2 to the/this stack frame, and restore.
	; XXX no need yet for one single type of frame.
	;
	; XXX actually, out_2 is in D register, if bbram is active

	ld a, (_out_2)
	out (OUT_2), a

	and # O2_ROM0 | O2_ROMA14
	cp  # O2_ROM0
	jr nz, oops     ; unexpected stuff in the mirror byte

	rst DISABLE     ; make things look regular

	push hl         ; return from nmi, sometimes happens
	jp _nmi

;----------------------------------------------------------------------

continue:             ; boot
	xor a
	ld i, a           ; flag for nmi (needed if a soft reset)
	ld r, a           ; R.7 could be another flag someday
	im 1              ; rst 0x38 "oops" if never happens

	out (DA_TPC), a   ; avoid disaster
	out (WDOG), a

	; fix things for contiguous 48 kB ROM and 16 kB work RAM

	ld a, # O2_SMEM | O2_ROM0
	out (OUT_2),a              ; P8N: work-ram. both: ROM0

	ld a, # 1
	out (CSMEM), a    ; P8E: work-ram

	; short settling time.
	; on L8TM, waits DS1210 to wake up
0$:
	xor a
1$:
	ld hl, # _data    ; scratch location in ram
	ld (hl), a
	cp (hl)
	jr nz, 0$         ; did not stick. restart loop
	dec a
	jr nz, 1$         ; test all 256 bit patterns (on this single byte)

	; stack usable now (hopefully).
	; reset interrupt logic of peripherals

	ld b, # 0         ; 256 times (dozen or so would do)
2$:
	rst RETI
	djnz 2$

	ld de, # _data    ; copy data
	ld hl, # _etext+2 ; from rom image, built by mkbin.c
	ld bc, (_etext)   ; 1st word is length
	ldir

	xor a             ; zero bss and stack
3$:
	ld (de), a
	inc de
	bit 7, d
	jr nz, 3$         ; upto and including FFFF

	call gsinit       ; argh

	ld a, (_out_1)    ; O1_TXOFF. power was dropped earlier
	out (OUT_1), a

	call logic_probe  ; P8E or P8N ?
	ld (_logic), a    ; P8E 0, P8N 1

	call _initialise  ; lotsa OUT etc

	ld a, # 1
	ld (ticket), a    ; allow tick()

	ld a, #>_iv
	im 2              ; ready now
	ld i, a           ; now nmi doesn't just boot

	call __enable
	jp mainloop       ; forever.

;----------------------------------------------------------------------
	.area _CODE      ; org 0x0100

	; does not belong here,
	; but where is .align ?

	.include "tables.inc"

;----------------------------------------------------------------------

mainloop:
	ld hl, # 0
	ld (_fido), hl   ; 10 second watchdog in tick()

	call __enable    ; here, in case it is botched somewhere
	call _hare       ; every ?

	ld de, (sluggish)
	ld hl, (_timef)
	ld a, e
	cp l
	call nz, 1$

	jp mainloop
1$:
	ld (sluggish), hl
	ld a, d
	cp h
	jp z, _tortoise  ; every 4 msec, except

	jp _slug         ; every second

;----------------------------------------------------------------------
; 82C54 CLK_1  4.032 MHz ~248 nsec
; P8N   clock  4.032 MHz
; P8E   clock  8.064 MHz ~124 nsec (1 waitstate per each opcode fetch)

logic_probe:

	ld c, # TMR_CTRL
	ld b, # TMR_SEL_MT | TMR_LATCH
	ld a, # TMR_SEL_MT | TMR_LSB | TMR_INTTC
	out (c), a

	ld a, # 73        ; near midpoint
	out (TMR_MT), a   ; count 73

	.byte 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 ; 20 nops, 100T or 80T

	out (c), b        ; latch count
	in a, (TMR_MT)    ; P8E count 18, P8N count -18
	rlca
	and # 1           ; P8E 0, P8N 1
	ret

;----------------------------------------------------------------------

to_int:
	ld a, # 1
	ld (DInest), a
	rst JPHL              ; handler()
	di                    ; might have allowed nesting
	rst RETI              ; some extra cycles this way, but ...

nmi_ei_ret:               ; were enabled, so DInest has to be 0
popa_ei_ret:
	pop iy                ; XXX place to switch tasks
	pop ix
	exx
	pop de
	pop bc
	pop af
	pop hl
	exx                   ; XXX double exx could be avoided
	ex af, af'
	pop de
	xor a
	ld (DInest), a
	pop bc
	pop af
	pop hl
	ei
	ret                   ; RETI, when needed, was done above

nmi_ret:
	ld hl, # DInest
	dec (hl)        ; restore DInest back to what it was
	jp m, _boot     ; out of sync, bug somewhere

	pop iy          ; oh where is .macro !
	pop ix
	exx             ; 2 copies of "popa" here
	pop de          ; always hl af bc de hl' af' bc' de' ix iy
	pop bc
	pop af
	pop hl
	exx
	ex af, af'
	pop de
	pop bc
	pop af
	pop hl          ; return without enabling
	ret

pusha:
	ex (sp), hl     ; only "pusha" is here, but look for pusha_some
	push af
	push bc
pusha_rest:
	push de
	ex af, af'
	exx             ; XXX double exx could be avoided
	push hl
	push af
	push bc
	push de
	exx
	push ix
	push iy
	jp (hl)         ; _boot, if bombstack active

fast_out:
	pop bc          ; reverse of "pusha_some"
	pop af
	pop hl
	ei
	reti

_swtch::
	call pusha
	di
	jp popa_ei_ret

;----------------------------------------------------------------------
; interrupts themselves.
; some routines do it all or partly here
;
; priority order is
; pa > pb > sa rx > sa tx > sa esc > sb rx > sb tx > sb esc
;----------------------------------------------------------------------
; 1968.75 Hz into PA.0
; slow things down to 256 Hz tick

adc_ret:                ; common part of pio_a_int (_XXX)
	pop hl
	ld (adc_reader), hl

;	ld hl, # - redstack
;	add hl, sp
;	jp nc, _boot        ; stack underflow (or bombstack)

	; increment 32.24 time

	ld bc, # 8522       ; 1.0 / 1968.75 * 2^24 ~ 1.0000 Hz
	ld hl, (_time_all)
	add hl, bc
	ld (_time_all), hl
	jp nc, fast_out

	ld hl, # _timef
	inc (hl)            ; seconds / 256
	jr nz, 1$
	inc l
	inc (hl)            ; seconds
	jr nz, 1$
	inc l
	inc (hl)
	jr nz, 1$
	inc l
	inc (hl)
	jr nz, 1$
	inc l
	inc (hl)            ; a rare event
1$:
	ld hl, # ticket
	srl (hl)
	jr nc, fast_out     ; previous tick has not finished (ominous)

	ld hl, # 2$
	jp pusha_rest       ; macros macros needed
2$:
	rst RETI            ; all clear to PIO
	ei
	call _tick

	; roll rssi history: 25, ... 8, 4 and 0 msec ago (or so).
	; ad_sql seems to have very long RC constant XXX

	ld hl, (_ad_rssi - 5)
	ld (_ad_rssi - 6), hl
	ld hl, (_ad_rssi - 3)
	ld (_ad_rssi - 4), hl
	ld hl, (_ad_rssi - 1)
	ld (_ad_rssi - 2), hl

	ld a, # 1           ; a valid ticket
	di
	ld (ticket), a

	jp popa_ei_ret


pio_a_int:
	push hl             ; pusha_some
	push af
	push bc

	ld hl, (adc_reader) ; one A/D finished, start another
	jp (hl)             ; returns at adc_ret:

pio_a_int_dcs::         ; 1968.75 Hz
	push hl             ; pusha_some
	push af
	push bc

    ld hl, (dcs_ptr)    ; 337 bytes makes it about 134.37 Hz
    ld a, (hl)
    out (DA_RFC), a
    inc hl
    bit 2, h            ; buffer ends at 0xFC00 1100
    jr z, 1$            ; valid area 0xFA?? 1010 or 0xFB?? 1011
    ld hl, # _dcs_wave  ; 0xFC00 - 337
1$:
    ld (dcs_ptr), hl

	ld hl, (adc_reader) ; one A/D finished, start another
	jp (hl)             ; returns at adc_ret:

pio_a_int_ctcss:
	push hl             ; pusha_some
	push af
	push bc

	ld bc, (ctcss_phinc)
	ld hl, (ctcss_phacc)
	add hl, bc
	ld (ctcss_phacc), hl

	ld l, h
	ld h, #>_ctcss_sine ; centered on RFC setting
	ld a, (hl)
	out (DA_RFC), a

	ld hl, (adc_reader) ; one A/D finished, start another
	jp (hl)             ; returns at adc_ret:

;----------------------------------------------------------------------

pio_b_int:
	call pusha
	ld hl, # _pio_b_int     ; could be a pointer variable
	jp to_int

;----------------------------------------------------------------------

sio_a_tbe:
	push hl
	push af
	ld hl, (_sio_a_tx_rem)
	ld a, (_sio_a_tx_ins)
	sub l                    ; get free zero
	jr z, 1$                 ; if no more
	ld a, (hl)               ; this probably happens more often
	out (SIO_A_DATA), a
	inc l
	ld (_sio_a_tx_rem), hl
	pop af
	pop hl
	ei
	reti
1$:
	ld (_sio_a_tx_active), a ; not anymore
	ld a, # WR0_RESET_TXINT
	out (SIO_A_CSR), a
	pop af
	pop hl
	ei
	reti

sio_a_rca:
	push hl
	push af
	ld hl, (_sio_a_rx_ins)
1$:
	in a, (SIO_A_DATA)
	ld (hl), a
	inc l
	in a, (SIO_A_CSR)         ; XXX most of the time not useful
	rra                       ; RR0_RCA to carry
	jr c, 1$
	ld (_sio_a_rx_ins), hl
	pop af
	pop hl
	ei
	reti

; parity, framing or overrun. just discard

sio_a_src:
	push af
	in a, (SIO_A_DATA)
	ld a, # WR0_ERROR_RESET
	out (SIO_A_CSR), a
	pop af
	ei
	reti

sio_a_esc:
	call pusha
	ld hl, # _sio_a_esc
	jp to_int

;----------------------------------------------------------------------

sio_b_tbe:
	call pusha
	ld hl, # _sio_b_tbe
	jp to_int

sio_b_rca_UNUSED:
	call pusha
	ld hl, # _sio_b_rca
	jp to_int

sio_b_rca:
	push hl
	push af
	ld hl, (_sio_b_rx_ins)
1$:
	in a, (SIO_B_DATA)
	ld (hl), a
	inc l
	in a, (SIO_B_CSR)
	rra                       ; RR0_RCA to carry
	jr c, 1$
	ld (_sio_b_rx_ins), hl
	pop af
	pop hl
	ei
	reti

sio_b_src:
	call pusha
	ld hl, # _sio_b_src
	jp to_int

sio_b_esc:
	call pusha
	ld hl, # _sio_b_esc
	jp to_int

;----------------------------------------------------------------------
; ADC0809
; ALE+START wired together
;
; sometimes, wiggling d/c power cable,
; ad_batt sticks at FF (ought to be around F0, 14+ volts)
;
; ordered, just for fun, to change only one mux bit
; before RSSI, SQL, BATT measurements.
; RSSI and SQL done double rate, always new values to tick.
;
; one straw still to try, try doing IN7 before BATT XXX

adc_first:
	out (AD_RSSI), a   ; leading edge ALE, channel in address bus lsbits
	nop                ; 2.5 usec from 1st ALE to 2nd START
	nop
	nop
adc_loop:
	out (AD_RSSI), a   ; leading edge SAR reset (ALE again), trailing START
	call adc_ret       ; resume later
	call adc_ret       ; might have been long disable(), every another time

	in a, (AD_RSSI)    ; OE (address bus don't care)
	out (AD_SQL), a
	ld (_ad_rssi), a    ; delay between OUTs, gratis
	out (AD_SQL), a
	call adc_ret
	call adc_ret

	in a, (AD_SQL)
	out (AD_IN7), a
	ld (_ad_sql), a
	out (AD_IN7), a
	call adc_ret
	call adc_ret

	in a, (AD_IN7)
	out (AD_TPC), a
	ld (_ad_in7), a
	out (AD_TPC), a
	call adc_ret
	call adc_ret

	in a, (AD_TPC)
	out (AD_BATT), a
	ld (_ad_tpc), a
	out (AD_BATT), a
	call adc_ret
	call adc_ret

	in a, (AD_BATT)
	out (AD_RSSI), a
	ld (_ad_batt), a
	out (AD_RSSI), a
	call adc_ret
	call adc_ret

	in a, (AD_RSSI)
	out (AD_SQL), a
	ld (_ad_rssi), a
	out (AD_SQL), a
	call adc_ret
	call adc_ret

	in a, (AD_SQL)
	out (AD_TP4), a
	ld (_ad_sql), a
	out (AD_TP4), a
	call adc_ret
	call adc_ret

	in a, (AD_TP4)
	out (AD_RPM), a
	ld (_ad_tp4), a
	out (AD_RPM), a
	call adc_ret
	call adc_ret

	in a, (AD_RPM)
	out (AD_FPM), a
	ld (_ad_rpm), a
	out (AD_FPM), a
	call adc_ret
	call adc_ret

	in a, (AD_FPM)
	out (AD_RSSI), a
	ld (_ad_fpm), a

	jp adc_loop

;----------------------------------------------------------------------
; most of the following routines must be in low rom, below 0x8000

_read_dtmf::
	pop de
	pop hl             ; 0 none, 1 MT8870, 2 MT8888/9
	push hl
	push de

	dec l
	jr z, _read_8870
	dec l
	jr z, _read_8888

	ld l, # 0          ; 'nothing' retval
	ret

_read_8888::

	rst DISABLE        ; get instantaneous state

	in a, (MT8888_RS)  ; read status
	cpl                ; DELAYED STEERING# in b3 is
	rla
	and # 0x10         ;  inverse of 8870 StD
	ld l, a

	in a, (MT8888_RD)  ; read data
	and # 0xF          ;  is valid before and after StD
	or l
	ld l, a

	jp _restore        ; "StD b3 b2 b1 b0"

_read_8870::           ; a byte from ROM1 (2nd one), the silly DTMF decoder

	ld hl, # _rom_bank ; into ROM1, any address, StD D3...0 ? ? afsqw

	rst DISABLE

	ld a, (_out_2)
	and # ~O2_ROM0   ; select ROM1. ROMA14 ... ROMA16 not touched
	out (OUT_2), a

	ld a, (hl)       ; if nmi comes and goes, might read ROM0[0x8000]
	and # 0xF8
	ld b, a

	ld a, # 0xF8
	and (hl)         ; spread the three (hl) accesses evenlyish
	cp b
	jp z, 1$         ; i'll take it
	ld a, # 0xF8
	and (hl)        ; must be good now XXX
1$:
	rrca
	rrca
	rrca             ; StD D3...0
	ld l, a          ; just the 8870 bits, rightjust

	ld a, (_out_2)
	out (OUT_2), a   ; back to regular state

	jp _restore

_read_ccir::
	in a, (PIO_A_DATA)
	and # PA_CCIR
1$:
	ld l, a
	in a, (PIO_A_DATA)
	and # PA_CCIR
	cp l
	jr nz, 1$        ; re-read while changing

	rrca
	rrca
	rrca
	rrca
	ld l, a          ; rightjust
	ret

;----------------------------------------------------------------------

_load_bbram::

	ld a, (_logic)   ; P8E 0, P8N 1, L8M 2, L8TM 3
	and # 3
	ret pe           ; P8E or L8TM, nothing to do

	rst DISABLE

	ld a, (_out_2)
	ld d, a          ; selects work ram
	and # ~O2_SMEM
	ld e, a          ; selects bbram
	ld c, # OUT_2

	ld hl, # _bbram  ; start
	ld a, #> _data   ; HI(end)

	ld (spsave), sp
	ld sp, # bombstack ; avoid corruption from nmi pushes. _boot if nmi
1$:
	out (c), e       ; bbram
	ld b, (hl)
	out (c), d       ; work ram back
	ld (hl), b
	inc hl
	cp h
	jp nz, 1$        ; _bbram ... _data

	ld sp, (spsave)  ; nmi not fatal anymore
	jp _restore      ; did it

_save_bbram::        ; almost identical XXX

	ld a, (_logic)
	and # 3
	ret pe           ; P8E or L8TM, nothing to do

	rst DISABLE

	ld a, (_out_2)
	ld d, a
	and # ~O2_SMEM
	ld e, a
	ld c, # OUT_2

	ld hl, # _bbram
	ld a, #> _data

	ld (spsave), sp
	ld sp, # bombstack
1$:
	ld b, (hl)
	out (c), e       ; bbram
	ld (hl), b
	out (c), d       ; work ram back
	inc hl
	cp h
	jp nz, 1$

	ld sp, (spsave)
	jp _restore

;----------------------------------------------------------------------

_sio_a_getc::             ; word sio_a_getc();
	ld hl, (_sio_a_rx_rem)
	ld a, (_sio_a_rx_ins)
	cp l
	jr nz, 1$
	ld hl, # -1       ; none == -1
	ret
1$:
	ld a, (hl)
	inc l
	ld (_sio_a_rx_rem), hl
	ld l, a
	ld h, # 0
	ret               ; zeroextended byte in HL

_sio_a_putc::         ; void sio_a_putc(word);
	pop hl
	pop bc            ; word arg
	push bc
	push hl
putc:
	rst DISABLE

	ld a, (_sio_a_tx_active)
	or a
	jp nz, 1$
	inc a
	ld (_sio_a_tx_active), a
	ld a, c
	out (SIO_A_DATA), a
	jp _restore          ; RET
1$:
	ld hl, (_sio_a_tx_ins)
	ld (hl), c
	inc l
	ld (_sio_a_tx_ins), hl
	jp _restore

_sio_a_puts::
	pop bc
	pop de
	push de
	push bc

	jr 2$
1$:
	ld c, a
	call putc
	inc de
2$:
	ld a, (de)
	or a
	jr nz, 1$
	ret

;----------------------------------------------------------------------

_ctcss_go::                   ; void ctcss_go(word phinc)
	pop de
	pop hl
	push hl
	push de

	ld (ctcss_phinc), hl         ; tone frequency

	ld hl, # pio_a_int_ctcss
	ld (_iv_pio_a), hl
	ret

_dcs_go::
	ld hl, # _dcs_wave
	ld (dcs_ptr), hl

	ld hl, # pio_a_int_dcs
	ld (_iv_pio_a), hl
	ret

_encoder_turnoff::

	ld a, (_iv_pio_a)            ; compare of lsbyte is enough
	cp # pio_a_int_dcs
	jr z, 1$
	cp # pio_a_int_ctcss
	ret nz                       ; huh?

	ld hl, # ctcss_phacc+1       ; CTCSS phase msbyte
	ld a, # 0x80                 ; msbit

	rst DISABLE

	xor (hl)
	ld (hl), a                   ; invert phase

	jp _restore
1$:
	ld hl, # pio_a_int           ; XXX should emit burst of 01010101010101
	ld (_iv_pio_a), hl           ; XXX at double speed
	ret

_encoder_end::

	ld hl, # pio_a_int           ; encoders off
	ld (_iv_pio_a), hl

	ld hl, # 0
	ld (ctcss_phacc), hl         ; next time, start from zero

	ret

;----------------------------------------------------------------------

_out_0_change::       ; void out_0_change(byte clear, byte set);
	pop hl
	pop de            ; e clear, d set
	push de
	push hl

	ld hl, # _out_0

	rst DISABLE

	ld a, (hl)        ; out_0 = (out_0 & ~clear) | set;
	cpl
	or e
	cpl
	or d
	ld (hl), a
	out (OUT_0),a
	jp _restore

_out_1_change::       ; void out_1_change(byte clear, byte set);
	pop hl
	pop de            ; e clear, d set
	push de
	push hl

	ld hl, # _out_1

	rst DISABLE

	ld a, (hl)        ; out_1 = (out_1 & ~clear) | set;
	cpl
	or e
	cpl
	or d
	ld (hl), a
	out (OUT_1),a
	jp _restore

_out_2_change::       ; void out_2_change(byte clear, byte set);
	pop hl
	pop de            ; e clear, d set
	push de
	push hl

	ld hl, # _out_2

	rst DISABLE

	ld a, (hl)        ; out_2 = (out_2 & ~clear) | set;
	cpl
	or e
	cpl
	or d
	ld (hl), a
	out (OUT_2),a
	jp _restore

_out_2_toggle::       ; void out_2_toggle(byte bits);
	pop hl
	pop de
	push de
	push hl

	ld a, e           ; bits
	ld hl, # _out_2

	rst DISABLE

	xor (hl)          ; out_2 ^= bits;
	ld (hl), a
	out (OUT_2), a
	jp _restore

out_2_set:            ; out_2_set bits from A

	ld hl, # _out_2

	rst DISABLE

	or (hl)           ; out_2 |= bits
	ld (hl), a
	out (OUT_2),a
	jp _restore

out_2_clr:            ; out_2_clr bits from A

	cpl               ; ~bits
	ld hl, # _out_2

	rst DISABLE

	and (hl)          ; out_2 &= ~bits
	ld (hl), a
	out (OUT_2),a
	jp _restore

;----------------------------------------------------------------------

_pio_a_change::       ; void pio_a_change(byte clear, byte set);
	pop hl
	pop de
	push de
	push hl

	ld hl, # _pio_a

	rst DISABLE

	ld a, (hl)        ; pio_a = (pio_a & ~clear) | set;
	cpl
	or e
	cpl
	or d
	ld (hl), a
	out (PIO_A_DATA), a
	jp _restore

_pio_b_change::       ; void pio_b_change(byte clear, byte set);
	pop hl
	pop de
	push de
	push hl

	ld hl, # _pio_b

	rst DISABLE

	ld a, (hl)        ; pio_b = (pio_b & ~clear) | set;
	cpl
	or e
	cpl
	or d
	ld (hl), a
	out (PIO_B_DATA), a
	jp _restore

_pio_a_change_ddr::   ; void pio_a_change_ddr(byte clear, byte set);
	pop hl
	pop de
	push de
	push hl

	ld hl, # _pio_a_ddr

	rst DISABLE

	ld a, (hl)        ; pio_b = (pio_b & ~clear) | set;
	cpl
	or e
	cpl
	or d
	ld (hl), a

	ld a, # 0xCF        ; ddr-set-command
	out (PIO_A_CTRL), a
	ld a, (hl)
	out (PIO_A_CTRL), a
	jp _restore

_pio_b_change_ddr::   ; void pio_b_change_ddr(byte clear, byte set);
	pop hl
	pop de
	push de
	push hl

	ld hl, # _pio_b_ddr

	rst DISABLE

	ld a, (hl)        ; pio_b = (pio_b & ~clear) | set;
	cpl
	or e
	cpl
	or d
	ld (hl), a

	ld a, # 0xCF        ; ddr-set-command
	out (PIO_B_CTRL), a
	ld a, (hl)
	out (PIO_B_CTRL), a
	jp _restore

;----------------------------------------------------------------------

_sio_a_change::       ; void sio_a_change(byte clear, byte set);
	pop hl
	pop de
	push de
	push hl

	ld hl, # _sio_a_wr5

	rst DISABLE

	ld a, (hl)        ; sio_a_wr5 = (sio_a_wr5 & ~clear) | set;
	cpl
	or e
	cpl
	or d
	ld (hl), a

	ld a, # 5
	out (SIO_A_CSR), a
	ld a, (hl)
	out (SIO_A_CSR), a
	jp _restore

_sio_b_change::       ; void sio_b_change(byte clear, byte set);
	pop hl
	pop de
	push de
	push hl

	ld hl, # _sio_b_wr5

	rst DISABLE

	ld a, (hl)        ; sio_b_wr5 = (sio_b_wr5 & ~clear) | set;
	cpl
	or e
	cpl
	or d
	ld (hl), a

	ld a, # 5
	out (SIO_B_CSR), a
	ld a, (hl)
	out (SIO_B_CSR), a
	jp _restore

_sio_b_toggle::       ; void sio_b_toggle(byte bits);
	pop hl
	pop de
	push de
	push hl

	ld a, e             ; bits
	ld hl, # _sio_b_wr5

	rst DISABLE

	xor (hl)          ; sio_b_wr5 ^= bits;
	ld (hl), a

	ld a, # 5
	out (SIO_B_CSR), a
	ld a, (hl)
	out (SIO_B_CSR), a
	jp _restore

;----------------------------------------------------------------------
; MB87006 or MC145156 or ...
;
; interrupts are disabled throughout pll_load()

_pll_bits::          ; void pll_bits(byte mask, byte bits);
	pop hl
	pop de           ; d bits, e mask
	push de
	push hl

	ld a, (_out_1)
	and #~(O1_SD | O1_CLK)     ; XXX O1_SRE, O1_STE, O1_SCE not touched
	ld (_out_1), a
	ld h, a          ; all those zero
2$:
	ld a, d
	and e
	jr z, 1$
	ld a, # O1_SD
1$:
	or h             ; TXON/OFF, TPS, RAS and SxE
	out (OUT_1), a
	or # O1_CLK
	out (OUT_1), a
	and #~O1_CLK
	out (OUT_1), a   ; 1 usec setup/hold/clock/anything

	srl e
	jp nz, 2$

	ret

;----------------------------------------------------------------------
; PCF2111T or 4035 shifters

_cu_bit::              ; void cu_bit(int zornot);
	pop hl
	pop de             ; int bit
	push de
	push hl

	ld a, e
	or d
	jr z, 1$           ; zerobit
	ld e, # O2_DP      ; one
1$:
	rst DISABLE

	ld a, (_out_2)
	and # ~(O2_CLK | O2_DP)
	or e

	out (OUT_2), a
	ld (_out_2), a     ; data setup/hold min 3 usec

	or # O2_CLK
	out (OUT_2), a

	and # ~O2_CLK      ; clock pulse min 1 usec
	out (OUT_2), a

	jp _restore        ; CLK zeroed, DP left 0 or 1

;----------------------------------------------------------------------
; SCL low on entry/exit
; SDA might or might not change

_i2c_bit::              ; byte(bool) i2c_bit(int zornot_out);
_i2c_wbit::
	pop de
	pop hl              ; output bit, 0 or nonzero int
	push hl
	push de

	ld a, l
	or h
	jr z, 1$
	ld h, # PB_DCU      ; this bit
1$:
wbit:
	ld a, # 0xCF        ; ddr-set-command

	rst DISABLE

	out (PIO_B_CTRL), a
	ld a, (_pio_b_ddr)  ; here, 1 means input (lucky this time)
	and # ~PB_DCU
	or h
	out (PIO_B_CTRL), a
	ld (_pio_b_ddr), a  ; change is permanent

	ld a, (_out_2)
	or # O2_CLK
	out (OUT_2), a      ; SCL high

	in a, (PIO_B_DATA)
	and # PB_DCU
	ld l, a             ; return bit

	ld a, (_out_2)
	out (OUT_2), a      ; SCL low

	jp _restore

; SCL low entry/exit
; SDA released on exit

_i2c_wbyte::            ; byte(bool) i2c_wbyte(byte);
	pop de
	pop bc              ; byte in c
	push bc
	push de

	ld b, # 8
	ld a, c
	rlca
	rlca
	rlca                ; msbit next to PB_DCU
1$:
	rlca
	ld c, a             ; keep in c
	and # PB_DCU
	ld h, a
	call wbit           ; h 0/PB_DCU
	ld a, l             ; return bit in l
	cp h
	jr nz, 2$           ; contested '1'
	ld a, c
	djnz 1$

	ld h, # PB_DCU
	jp wbit             ; return ack 0, nak PB_DCU
2$:
	ld l, # 1           ; return 1 if contest
	ret

; SDA must have been released

_i2c_rbit::             ; byte(bool) i2c_rbit();

	rst DISABLE

	ld a, (_out_2)
	or # O2_CLK
	out (OUT_2), a      ; SCL high

	in a, (PIO_B_DATA)
	and # PB_DCU
	ld l, a             ; return bit

	ld a, (_out_2)
	out (OUT_2), a      ; SCL low

	jp _restore

;----------------------------------------------------------------------
; XXX hopeless things, not even tried yet

	.IF 0

_dp_putc::
	pop de
	pop hl              ; byte in L
	push hl
	push de

	ld b, # 10          ; start, 8 data, stop
	ld h, # 0xFF        ; stopbits
	add hl, hl          ; prepend startbit

	rst DISABLE         ; XXX
1$:
	ld a, (_logic)
	cp # 0              ; P8E ?
	call z, 2$          ; burn some more

	ld a, (_out_2)      ; 0x80 is O2_DP
	rla                 ; make room
	rr h
	rr l
	rra                 ; suck it in
	ld (_out_2), a      ; new bit in mirror
	out (OUT_2), a      ; and in pin
	djnz 1$

	jp _restore         ; many many stopbits during overhead

2$:
	ret

; assume coming from pio_b_int

_dcu_getc::

	ld bc, # 0xFF00 + PB_DCU
	ld l, # 0x80
1$:
	ld a, (_logic)
	cp # 0              ; P8E ?
	call z, 2$          ; burn some more

	in a, (PIO_B_DATA)  ; PB_DCU = 0x08, slow to rotate into carry
	and c
	add b
	rr l
	jp nc, 1$

	ld a, # 0x97        ; latter falling edges in the
	out (PIO_B_CTRL), a ; character triggered another interrupt.
	ld a, # ~PB_DCU     ; flush it.
	out (PIO_B_CTRL), a ; XXX works this way ?

	ret

2$:
	ret

;----------------------------------------------------------------------

_cointoss::
	ld a, r
	and # 1
	ld l, a
	ret

	.ENDIF

;----------------------------------------------------------------------

	.area _GSINIT
gsinit:              ; SDCC collects initialisations here

;----------------------------------------------------------------------
	.area _DATA      ; org 0xD000
_data::

_iv:                 ; aligned
	.word 0,0,0,0
	.word 0,0,0,0
_iv_sio::
	.word sio_b_tbe  ; could as well keep it in rom,
	.word sio_b_esc  ; but for easy vector changes
	.word sio_b_rca
	.word sio_b_src
	.word sio_a_tbe
	.word sio_a_esc
	.word sio_a_rca
	.word sio_a_src
_iv_pio_a::
	.word pio_a_int
_iv_pio_b::
	.word pio_b_int  ; 10 words used

	.word 0          ; safety first,
	.word 0          ; reboot if never happens.
	.word 0
	.word 0
	.word 0
	.word 0
	.word 0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0
	.word 0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0
	.word 0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0
	.word 0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0
	.word 0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0
	.word 0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0
	.word 0,0,0,0, 0,0,0,0
	.word 0          ; 129 words, last at 0xD100

DInest::
	.byte 1          ; reset implies DI

ticket:
	.byte 0          ; 1 if ok to enter tick()

_time_all::          ; these together, within a page
	.byte 0
_time_frac::
	.byte 0
_timef::
	.byte 0
_timex::
_time::
	.word 0
	.word 0          ; 32.24 time, seconds

adc_reader:
	.word adc_first

break:
	.word _end

_sio_a_rx_ins:
	.word _sio_a_rx_buf
_sio_a_rx_rem:
	.word _sio_a_rx_buf

_sio_a_tx_ins:
	.word _sio_a_tx_buf
_sio_a_tx_rem:
	.word _sio_a_tx_buf
_sio_a_tx_active:
	.byte 0

_sio_b_rx_ins::
	.word _sio_b_rx_buf
_sio_b_rx_rem::
	.word _sio_b_rx_buf

_sio_b_tx_ins::
	.word _sio_b_tx_buf
_sio_b_tx_rem::
	.word _sio_b_tx_buf
_sio_b_tx_active::
	.byte 0

;----------------------------------------------------------------------
	.area _BSS
bss::

_logic::   .ds 1       ; P8E, P8N, L8M, L8TM ?
spsave:    .ds 2       ; see save_bbram
sluggish:  .ds 2       ; tortoise/slug slowdown

_ad_rssi_25ms_ago::
             .ds 6     ; some older values ...
_ad_rssi::   .ds 1     ; ... and last. keep these seven together
_ad_sql::    .ds 1     ; ad_xxx in channel order (just for fun)
_ad_batt::   .ds 1
_ad_tpc::    .ds 1
_ad_fpm::    .ds 1
_ad_rpm::    .ds 1
_ad_tp4::    .ds 1
_ad_in7::    .ds 1

ctcss_phacc: .ds 2
ctcss_phinc: .ds 2

dcs_ptr:     .ds 2     ; into dcs_wave[337]

;----------------------------------------------------------------------
