
#include "head.h"
#include "cu.h"
#include "cfg.h"
#include "iomap.h"

/*
 *  F for finnish in CU58AF ?
 *
 *  PCF8576
 *  PCF8574
 *  PCF8574A
 *  PCD3312
 *
 *  XXX handset splits in two: DM58 and HS58.
 *  XXX support just DM58 ?
 *
 *     76543210
 *    876543210
 *
 *  ( 0   2) 4  6  8 10 12 14 16 18
 *  (20) 22 24 26 28 30 32 34 36 38
 *
 *  icon byte at offset 3,
 *  top row at offset 4,
 *  bottom row at offset 22,
 *  in the full 40 byte blast.
 *
 *  -----       3
 *  |\|/|   6 2 E F B
 *  -- --     5   A
 *  |/|\|   4 0 1 D 9
 *  -----       C
 */

static const word font[128] = {
	0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0,
	0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0,
	0x0000,   /*   */
	0x918C,   /* ! */
	0x41C0,   /* " */
	0x17B0,   /* # */
	0x57EA,   /* $ */
	0x83C1,   /* % */
	0xB18D,   /* & */
	0x01C0,   /* ' */
	0xA180,   /* ( */
	0x0185,   /* ) */
	0xE5A7,   /* * */
	0x45A2,   /* + */
	0x0181,   /* , */
	0x05A0,   /* - */
	0x0190,   /* . */
	0x8181,   /* / */
	0x9A59,   /* 0 */
	0x4002,   /* 1 */
	0x1C38,   /* 2 */
	0x1E28,   /* 3 */
	0x0E60,   /* 4 */
	0x1668,   /* 5 */
	0x1678,   /* 6 */
	0x0A08,   /* 7 */
	0x1E78,   /* 8 */
	0x1E68,   /* 9 */
	0x01D0,   /* : */
	0x4181,   /* ; */
	0xA180,   /* < */
	0x15A0,   /* = */
	0x0185,   /* > */
	0x818A,   /* ? */
	0x5DD8,   /* @ */
	0x0E78,   /* A */
	0xB078,   /* B */
	0x1058,   /* C */
	0x5A0A,   /* D */
	0x1478,   /* E */
	0x0478,   /* F */
	0x1658,   /* G */
	0x0FF0,   /* H */
	0x518A,   /* I */
	0x1B90,   /* J */
	0xA1F0,   /* K */
	0x11D0,   /* L */
	0x8BD4,   /* M */
	0x2BD4,   /* N */
	0x1BD8,   /* O */
	0x0DF8,   /* P */
	0x3BD8,   /* Q */
	0x2DF8,   /* R */
	0x318C,   /* S */
	0x418A,   /* T */
	0x1BD0,   /* U */
	0x81D1,   /* V */
	0x2BD1,   /* W */
	0xA185,   /* X */
	0x8186,   /* Y */
	0x9189,   /* Z */
	0x0180,   /* [ */
	0x2184,   /* \ */
	0x0180,   /* ] */
	0x0180,   /* ^ */
	0x1180,   /* _ */
	0x0184,   /* ` */
	0x0E78,   /* a */
	0xB078,   /* b */
	0x1058,   /* c */
	0x5A0A,   /* d */
	0x1478,   /* e */
	0x0478,   /* f */
	0x1658,   /* g */
	0x0FF0,   /* h */
	0x518A,   /* i */
	0x1B90,   /* j */
	0xA1F0,   /* k */
	0x11D0,   /* l */
	0x8BD4,   /* m */
	0x2BD4,   /* n */
	0x1BD8,   /* o */
	0x0DF8,   /* p */
	0x3BD8,   /* q */
	0x2DF8,   /* r */
	0x318C,   /* s */
	0x418A,   /* t */
	0x1BD0,   /* u */
	0x81D1,   /* v */
	0x2BD1,   /* w */
	0xA185,   /* x */
	0x8186,   /* y */
	0x9189,   /* z */
	0x0180,   /* { */
	0x4182,   /* | */
	0x0180,   /* } */
	0x0188,   /* ~ */
	0x0180,   /* DEL */
};

#define LCD        0x70 /* DM58 */
#define LEDS       0x44 /* DM58 */
#define DTMF       0x48 /* HS58 */

#define ROWS       0x4C /* DM58 */
#define ROW_MASK   0x7F /* bit7 no connect */

#define COLS       0x40 /* both modules. two chips, same address */
#define COL_MASK   0x07
#define BTN_MASK   0xF8 /* bit3 label "POWER", but function ? */
#define LDR        0x10
#define SPEAKER    0x20 /* side button */
#define TANGENT    0x40 /* PTT */
#define OFFHOOK    0x80

#define CTRL       0x7C /* HS58 */
#define SPK_OFF    0x01
#define DTMF_ON    0x02
#define CONNECT_AF 0x04 /* handset <> phone */
#define MIC_ON     0x08
#define EAR_ON     0x10
#define VOL_MASK   0xE0 /* 8 levels, lowest audible too */
#define VOL_LSB    0x20

/* output only scl */

#define scl_high()  out_2_set(O2_CLK)
#define scl_low()   out_2_clr(O2_CLK)

/* XXX might read back and verify written bits some day */

#define sda_release() pio_b_input(PB_DCU)
#define sda_pull()    pio_b_output(PB_DCU)

#define sda_read() (ior(PIO_B_DATA) & PB_DCU)
#define int_read() ((sio_a_rr0 & SA_DA) == 0) /* inversion by sio */

extern byte sio_a_rr0;

extern byte
	i2c_wbyte(byte),
	i2c_wbit(int),
	i2c_rbit();

static byte
i2c_rbyte(int nak)
{
	byte mask, data = 0;

	sda_release();

	for (mask = 0x80; mask; mask >>= 1)
		if (i2c_rbit())
			data |= mask;

	i2c_wbit(nak);

	return (data);
}

/* on entry, SCL low */

static void
i2c_stop()
{
	sda_pull();
	scl_high();
	sda_release();
}

static byte
i2c_start(byte addr)
{
	byte try = 10;

	scl_high();
	sda_release();

	/* confusion ? try to unstuck */

	while (sda_read() == 0) {

		if (--try == 0)
			return (1);      /* nz is error (DCU stuck low ?) */

		scl_low();
		scl_high();
	}
	sda_pull();
	scl_low();

	return (i2c_wbyte(addr));  /* 0 is ok */
}

static byte
i2c_write(byte addr, byte n, byte *p)
{
	if (i2c_start(addr) == 0) {
		do {
			if (i2c_wbyte(*p++))
				break;
		} while (--n);
	}
	i2c_stop();

	return (n);       /* 0 is ok */
}

static byte
i2c_put(byte addr, byte x)
{
	byte error = i2c_start(addr) || i2c_wbyte(x);

	i2c_stop();

	return (error);
}

static byte
i2c_get(byte addr)
{
	byte x = 0xFF;                   /* fake an inactive 8574 */

	if (i2c_start(addr | 0x01) == 0)
		x = i2c_rbyte(1);            /* send nak */

	i2c_stop();

	return (x);
}

/* XXX do piecemeal instead
 * and/or keep copy of display contents (first/last changed) */

static void
cu58af_update()
{
	byte x[2 + 1 + 2 * (8 + 1 + 9)];
	word *w;
	byte ico, n, *d;

	/* mimic CU53 */

	ico = 0;
	if (icon_star)
		ico |= 0x80; /* chevron 4 */
	if (icon_mast)
		ico |= 0x40; /* chevron 3 */
	if (icon_v_u)
		ico |= 0x20; /* chevron 2 */
	if (icon_v_d)
		ico |= 0x10; /* chevron 1 */
	if (icon_phone)
		ico |= 0x08;
	if (icon_car_u | icon_car_d)
		ico |= 0x04;
	if (0)
		ico |= 0x02; /* ALPHA */
	if (0)
		ico |= 0x01; /* EXP */

	x[0] = 0x86; /* C, pointer. skip unused area XXX 6 ? */
	x[1] = 0x60; /* device 0 */

	x[2] = ico;
	w = (word *) &x[3]; /* character area */

	d = display_1;
	n = 8;
	do {
		*w++ = font[*d++];
	} while (--n);

	*w++ = 0; /* unused slot between rows */

	d = keycnt ? buffer : display_2;
	n = 9;
	do {
		*w++ = font[*d++];
	} while (--n);

	i2c_write(LCD, sizeof(x), x);
}

static void
cu58af_leds()
{
	byte x = 0;

	if (led_lcdlight)
		x |= 0x40;
	if (led_keylight)
		x |= 0x20;

	if (led_roam)
		x |= 0x08;
	if (led_on)
		x |= 0x04;
	if (led_call)
		x |= 0x02;
	if (led_serv)
		x |= 0x01;

	i2c_put(LEDS, x);
}

/*
 *  B
 *
 *  STO RCL CL ENT    1    2   3
 *                   ABC  DEF GHI
 *
 *                    4    5   6
 *                   JKL  MNO PQR
 *
 *                    7    8   9
 *                   STU  VWX _YZ
 *
 *                    *    0   #
 *             +  -
 *
 *  two crosspoints empty, marked ? and Z
 */

static const byte keymap[] = "?ZB421563789*0#-CE+SR";

static const byte cu58af_alphakeys[] =
	"1ABC12DEF23GHI3"
	"4JKL45MNO56PQR6"
	"7STU78VWX89 YZ9"
	"**0-/,.?0##";

byte
	cu58af_buttons; /* positive logic here */

/* anything else than ROW_ or COL_MASK is kept '1',
 * and cannot be used for anything.
 * there are other pins in the handset to play with */

static byte
cu58af_keypad()
{
	byte r, c, key = 0;           /* 0 means no keys are pressed */

	if (int_read())
		return (0xFF);            /* /INT high, nothing changed here */

	c = ~i2c_get(COLS);           /* use positive logic */

	cu58af_buttons = c & BTN_MASK;

	if (c & COL_MASK) {

		/* rows high, cols low.
		 * retry while transfer error (until 10 second watchdog) */

		while (i2c_put(ROWS, ~0))
			;
		while (i2c_put(COLS, ~COL_MASK))
			;

		r = ~i2c_get(ROWS);

		/* restore cols input, rows output */

		while (i2c_put(ROWS, ~ROW_MASK))
			;
		while (i2c_put(COLS, ~0))
			;

		/* read(,flip,read) would have been nice, but no /INT from ROWS */

		c = ~i2c_get(COLS);         /* flush /INT (necessary ?) */

		cu58af_buttons = c & BTN_MASK;

		c &= COL_MASK;
		r &= ROW_MASK;

		if (r && c) {
			while ((r & 1) == 0) {
				r >>= 1;
				key += 3;
			}
			while ((c & 1) == 0) {
				c >>= 1;
				key += 1;
			}
			if ((r | c) & 0xFE)      /* multiple rows or cols ? */
				key = 0xFF;
			else
				key = keymap[key];
		}
	}
	/* 'shiftlike' */

	if (cu58af_buttons & SPEAKER)
		lights_on();

	return (key);
}

static void
cu58af_volume()
{
	byte x = CONNECT_AF | MIC_ON;

	if (cfg.volume == 0 || cfg.af_dest != 2)
		x |= SPK_OFF;

	if (cfg.volume == 1 || cfg.af_dest == 0)
		x |= EAR_ON;

	if (cfg.volume >= 2)
		x |= ((cfg.volume - 2) * VOL_LSB) & VOL_MASK;

	i2c_put(CTRL, x);
}

static void
cu58af_dtmf(byte c)
{
	byte x = 0;

	if (c) {
		if (isdigit(c)) {
			x = (c - '0') | 0x10;
		} else {
			switch (c) {
			case 'S': x = 0x1A; break;
			case 'R': x = 0x1B; break;
			case 'C': x = 0x1C; break;
			case 'E': x = 0x1D; break;
			case '*': x = 0x1E; break;
			case '#': x = 0x1F; break;

			case 'B': x = 0x3F; break; /* 1768.5 Hz */
			}
		}
	}
	if (x) {
		i2c_put(DTMF, x);

		x = CONNECT_AF | VOL_MASK | 0 | DTMF_ON | EAR_ON | SPK_OFF;
		i2c_put(CTRL, x);
	} else {
		i2c_put(DTMF, 0x01); /* no-tone, osc OFF */

		cu58af_volume();     /* restore */
	}
}

/* when HS58 is missing */

static void
cu58af_dtmf_2(byte c)
{
	if (c == 0)
		return;

	if (isdigit(c)) {
		c -= '0';
	} else {
		switch (c) {
		case 'S': c = 10; break;
		case 'R': c = 11; break;
		case 'C': c = 12; break;
		case 'E': c = 13; break;
		case '*': c = 14; break;
		case '#': c = 15; break;

		default:  c = 16; break; /* 1768.5 Hz */
		}
	}
	make_sine(dtmf_sine, cfg.dtmf_gain, 0x40);

	timer_mt(0);
	out_0_set(O0_CCIRC | O0_MTC | O0_MICM);

	/* busywaits /INT (photoresistor can change XXX) */

	dtmf_timer_mt_b(dtmf_mt_tab[c].phinc_1,
	                dtmf_mt_tab[c].phinc_2);

	out_0_clr(O0_CCIRC | O0_MTC | O0_MICM);
	timer_mt(0);
}

byte
cu58af_probe()
{
	byte x[5 + 40];

	bzero(x + 5, 40);

	x[0] = 0xCC; /* C, normal mode, enabled, 1/2 bias, 4 backplanes */
	x[1] = 0x80; /* C, data pointer rewind */
	x[2] = 0xE0; /* C, device select 0 */
	x[3] = 0xF8; /* C, bank select 0 */
	x[4] = 0x70; /* blink off */

	i2c_put(DTMF, 0x01); /* first transfer might fail */

	/* all these 3 chips are in DM58 */

	if (i2c_write(LCD, sizeof(x), x)
	 || i2c_put(ROWS, ~ROW_MASK)
	 || i2c_put(COLS, ~0))
		return (0);

	cu58af_dtmf(0); /* sets volume too */
	cu58af_leds();
	cu58af_keypad();

	display_1_len = 8;
	display_2_len = 9;

	alphakeys = cu58af_alphakeys;

	cu_keypad = cu58af_keypad;
	cu_update = cu58af_update;
	cu_leds   = cu58af_leds;

	/* HS58 might not exist */

	if (i2c_put(DTMF, 0x01)) {
		cu_dtmf   = cu58af_dtmf_2; /* have to bang it. no volume either. */
	} else {
		cu_dtmf   = cu58af_dtmf;
		cu_volume = cu58af_volume;
	}
	return (1);
}

