
#ifdef SCCZ80    /* is Z88DK */
#define volatile /* nothing */
#define const    /* nothing */
#endif

typedef unsigned char  byte;
typedef unsigned short word;
typedef unsigned long  dword;

#define NULL      ((void *) 0)
#define __SIZE(r) (sizeof(r)/sizeof(*(r)))

#define isprint(c) ((byte)((c)-' ') < 96)
#define isdigit(c) ((byte)((c)-'0') < 10)
#define isupper(c) ((byte)((c)-'A') < 'Z'-'A'+1)

#define offsetof(t,m) ((word) &((t *)0)->m)

extern void
	boot(),
	power_off(),

	disable(),
	restore(),
	enable(),

	load_bbram(),
	save_bbram(),

	ctcss(word dHz),
	ctcss_go(word phinc),
	dcs(int code),
	dcs_go(),
	encoder_turnoff(),
	encoder_end(),

	pll_bits(byte mask, byte bits),
	cu_bit(int zornot),

	axaddr(byte *dst, byte *asc),
	axout(byte *, word len, word idle),
	lights_on(),
	lights_off(),
	nop();

extern dword
	dcs_frame(int code);

extern byte
    i2c_bit(word zornot),
	read_ccir(),
	read_dtmf(byte type),
	read_8870(),
	read_8888(),
	cointoss();            /* randomish 0 or 1 */

extern const byte
	version[];

word strlen  (byte *);
byte *strend (byte *);
byte *index  (byte *, word c);
byte *rindex (byte *, word c);

int strcmp   (byte *, byte *);

byte *strcpyx  (byte *dst, byte *src);
byte *strncpyx (byte *dst, byte *src, word bufsiz);

void bcopy (void *src, void *dst, word len);
void bzero (void *, word len);
byte bcmp  (void *, void *, word len);

void memcpy (void *dst, void *src, word len);
void memset (void *dst, word c, word len);

word crc_ccitt(byte *, word len);

#define delay_msec(ms) delay_frac(((ms) * 65536) / 1000)

void delay_frac(word of_a_second);  /* uses 'time' */
void delay_usec(word usec);         /* uses 'logic' for cyclecount */

void store_word(word *, word);

void make_sine(byte *, byte gain, byte center);

word sio_a_getc();
void sio_a_putc(word);
void sio_a_puts(byte *);

extern byte hexdig[];

byte nybble(char);

byte  atob (byte *);
word  atow (byte *);
dword atodw(byte *);

byte
	ulmoddiv_10(dword *inout),
	ulmoddiv_91(dword *inout),
	ulmoddiv(dword *, byte);

byte
	umoddiv_10(word *inout),
	umoddiv_91(word *inout),
	umoddiv(word *, byte);

word
	abs(int),
	umulbb(byte, byte);

dword
	labs(long);

/* different views into 32.24 time, seconds */

extern volatile dword timex;      /* seconds, 32 bit */
extern volatile word  time;       /* seconds, 16 bit */
extern volatile word  timef;      /* seconds, 8.8    */
extern volatile word  time_frac;  /* seconds, 0.16   */

extern dword timex_get();

extern word timef_get();
extern word timef_since(word gotten);

extern word timef_mark(word interval);
extern byte timef_passed(word mark);

extern word time_mark(word interval);
extern byte time_passed(word mark);

/* enumeration order is important */

extern byte logic;
enum {
	P8E,   /* only 8 MHz */
	P8N,   /* cfg must be moved to non-volatile ram area */
	L8M,
	L8TM,  /* & 3 == 3 */
};

extern byte
	ad_rssi_25ms_ago,
	ad_rssi,
	ad_sql,
	ad_batt,
	ad_tpc,
	ad_fpm,
	ad_rpm,
	ad_tp4,
	ad_in7,

	receiving,
	transmitting,
	in_setup;

extern word
	debug;

extern dword
	rx_freq,
	tx_freq;

byte RFC_get();
void RFC_set(byte rfc);

enum { FSK, CCIR, DTMF }; /* buf type */

struct buf {
	struct buf *next;
	byte len;
	byte type;
	byte data[4 + 13 + 2]; /* some preamble, sync word, data, crc */
};

extern struct buf
	*fx_tx_head, **fx_tx_tail,
	*incoming_head, **incoming_tail,
	*buf_free,
	*buf_alloc();

extern void
	buf_relse(struct buf *),
	nmea_input(byte);

extern int
	nmea_split(byte *);

extern void
	dtmf_timer_mt(word, word),
	dtmf_timer_mt_b(word, word);

extern
struct dtmf_mt_tab {
	word phinc_1, phinc_2;
}
	const dtmf_mt_tab[];

extern byte dtmf_sine[];

