
/*
 *  input intel hex
 *
 *     0 ... BFFF code
 *  C000 ... CFFF forbidden zone
 *  D000 ... FFFF data
 *
 *  output binary image, code and data appended
 *  filled with FFh
 *
 *  XXX when 48 kB is exceeded, fix boot.s to
 *  XXX read data image from 2nd rom bank (ROMA14)
 */

#define ROM_END 0xC000
#define DATA    0xD000

#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <strings.h>
#include <string.h>
#include <stdio.h>

static u_char file[0x10000];
static u_long
	etext,
	edata;

static int
file_chunk(u_char *buf, u_long count, u_long addr)
{
	u_long e = addr + count;

	if (e < addr
	 || e > sizeof(file))
		return (-1);

	/* C000 ... D000 ? */

	if (addr < ROM_END && e >= DATA)
		return (-1);

	if (addr >= ROM_END && addr < DATA)
		return (-1);

	if (e >= ROM_END && e < DATA)
		return (-1);

	if (addr < ROM_END && e > etext)
		etext = e;

	if (addr >= DATA && e > edata)
		edata = e;

	if (count)
		bcopy(buf, &file[addr], count);

	return (0);
}

static int
anybble()
{
	int c = getchar();

	if (c >= '0' && c <= '9')
		return (c - '0');

	if (c >= 'A' && c <= 'F')
		return (c - 'A' + 10);

	if (c >= 'a' && c <= 'f')
		return (c - 'a' + 10);

	return (-1);
}

static u_char cksum;

static int
abyte(u_char *p)
{
	int
		a = anybble(),
		b = anybble();

	if (a == -1 || b == -1)
		return (-1);

	cksum += *p = (a << 4) | b;

	return (0);
}

load_intel()
{
	u_char count, offhi, offlo, type, buf[256 + 1];
	u_long base = 0;
	int i;

	while ((i = getchar()) != EOF) {
		if (isspace(i))
			continue;

		cksum = 0;

		if (i != ':'
		 || abyte(&count)
		 || abyte(&offhi)
		 || abyte(&offlo)
		 || abyte(&type))
			goto bad;

		/* include checksum byte */

		for (i = 0; i < count + 1; ++i)
			if (abyte(&buf[i]))
				goto bad;
		if (cksum)
			goto bad;

		switch (type) {

		case 0x00:
			if (file_chunk(buf, count, base + ((offhi << 8) | offlo)))
				return (-1);
			break;

		case 0x02:
			if (count != 2)
				goto bad;
			base = ((buf[0] << 8) | buf[1]) << 4;
			break;

		case 0x04:
			if (count != 2)
				goto bad;
			base = ((buf[0] << 8) | buf[1]) << 16;
			break;

		case 0x03:
		case 0x05:
		case 0x01:
			break;

		default:
	bad:
			return (-1);
		}
	}
	return (0);
}

main(int argc, char **argv)
{
	u_long datalen;

	memset(file, 0xFF, sizeof(file));

	if (load_intel()) {
		fprintf(stderr, "bad hex\n");
		exit(1);
	}
	if (etext + 2 + datalen > ROM_END) {
		fprintf(stderr, "too much\n");
		exit(1);
	}
	if (isatty(1)) {
		fprintf(stderr, "not to tty\n");
		exit(1);
	}

	fprintf(stderr, "etext 0x%X %5d\n", etext, etext);
	fprintf(stderr, "edata 0x%X %5d\n", edata, edata);

	datalen = edata - DATA;

	file[etext + 0] = datalen;
	file[etext + 1] = datalen >> 8;

	bcopy(&file[DATA], &file[etext + 2], datalen);

	write(1, file, ROM_END);

	return (0);
}

