
#include "head.h"
#include "cfg.h"

static byte
	nmea_len = 0xFF, /* $ needed */
	nmea_buf[128];

typedef byte *byte_ptr;

union nmea_sentence {

	byte_ptr fields[16];

	/* std */

	struct RMC { byte_ptr
		utc,
		validity,
		lat, lat_hem,
		lon, lon_hem,
		knots, course,
		date,
		magnetic_variation, magnetic_variation_hem,
		FAA_indicator;
	} RMC;

	struct RMB { byte_ptr
		validity,
		off_track, steer,
		last_wpt, next_wpt,
		lat, lat_hem,       /* of next_wpt */
		lon, lon_hem,
		range, bearing, closing_velocity, arrival_alarm;
	} RMB;

	struct GGA { byte_ptr
		utc,
		lat, lat_hem,
		lon, lon_hem,
		quality, nsat, HDOP,
		altitude, altitude_unit,
		geoid_height, geoid_height_unit,
		time_since_DGPS, DGPS_station_id;
	} GGA;

	struct GLL { byte_ptr
		lat, lat_hem,
		lon, lon_hem,
		utc,
		validity,
		FAA_indicator;
	} GLL;

	struct ZDA { byte_ptr
		utc,
		day, month, year;
	} ZDA;

	struct VTG { byte_ptr
		true_course, T, magnetic_course, M,
		knots, N, kmh, K;
	} VTG;

	struct GSA { byte_ptr
		operation_mode,
		fix_dimension,
		satellite_PRN[12],
		PDOP, HDOP, VDOP;
	} GSA;

	struct GSV { byte_ptr
		gsv_count, gsv_index, /* first of two looks like 2, 1 */
		nsat_in_view;
		struct { byte_ptr
			PRN,
			elevation, azimuth, SNR;
		} sats[4];
	} GSV;

	struct AAM { byte_ptr
		circle_entered, perpendicular_passed,
		circle_radius, circle_unit,
		wpt_name;
	} AAM;

	/* private */

	struct PMGNST { byte_ptr
		version;
	} PMGNST;

	struct PGRMZ { byte_ptr
		altitude, altitude_unit, altimeter_source;
	} PGRMZ;

	struct PGRME { byte_ptr
		horizontal, horizontal_unit,
		vertical,   vertical_unit,
		spherical,  spherical_unit;
	} PGRME;

	struct PGRMM { byte_ptr
		datum;
	} PGRMM;

	struct PSLIB { byte_ptr
		kHz, baud,
		request;
	} PSLIB;

	struct PSNY { byte_ptr
		preamp_status;
	} PSNY;

	struct PEMT { byte_ptr
		command;
		union {
			byte_ptr unknown;
		} u;
	} PEMT;

	struct PRAYA { byte_ptr
		unknown;
	} PRAYA;
}
	nmea; /* filled by nmea_split() */

static void
process_nmea()
{
	int i;

	sio_a_puts(nmea_buf);

	for (i = 0; i < __SIZE(nmea.fields); ++i) {
		if (nmea.fields[i] == NULL)
			break;
		sio_a_putc('-');
		sio_a_puts(nmea.fields[i]);
	}
	sio_a_putc('\n');
}

void
nmea_input(byte c)
{
	if (c < ' ') {
		if ((c == 0x0A || c == 0x0D)
		 && nmea_len < __SIZE(nmea_buf) - 1
		 && nmea_len > 10
		 && isupper(nmea_buf[0])
		 && isupper(nmea_buf[1])) {
			nmea_buf[nmea_len] = 0;

			if (nmea_split(nmea_buf) == 0)
				process_nmea();
		}
		nmea_len = 0xFF; /* $ needed */

	} else if (c == '$') {
		nmea_len = 0;

	} else if (nmea_len < __SIZE(nmea_buf) - 1)
		nmea_buf[nmea_len++] = c;
}

