
#define STUFF /* hide extern */

#include "head.h"
#include "cfg.h"
#include "cu.h"
#include "iomap.h"

const byte version[] = "@(#)R58bis 080814#22"; /* YYMMDD#rev */

const word cfg_size = sizeof(cfg);

#define TCXO    12800000U
#define TMR_CLK  4032000U

#define TICK_HZ      256U
#define MSEC       * 256L / 1000 /* for timef_mark() et al */

/* CPU clock can be 8.064 or 4.032 MHz,
 * logic == 0 == P8E, if 8.064 MHz, all others 4.032 MHz */

/* i/o mirrors and misc */

byte
	/* P8x */

	out_0 = O0_INH | O0_MICM,
	out_1 = O1_TXOFF,
	out_2 = O2_SMEM | O2_ROM0 | O2_CS2 | O2_CS1,

	pio_a,
	pio_a_ddr = 0xFF,         /* 1 input, 0 output pin */
	pio_b,
	pio_b_ddr = PB_TMR0 | PB_RXON | PB_DCU | PB_EXIN2 | PB_EXIN1,

	sio_a_wr5 = WR5_8BIT_TX | WR5_TX_ENB,
	sio_a_rr0,
	sio_b_wr5 = WR5_8BIT_TX | WR5_TX_ENB,
	sio_b_rr0,

	fx429_cmd,

	/* generic */

	squelch_open,
	squelch_timer,
	audio_flags,       /* audio muted or opened somehow */
	tx_vco_on,
	ptt_works,
	last_memory = -1,
	transmitting;

word
	dtmf_timo,
	batt_timo,
	debug,
	fido;       /* reboot, unless zeroed periodically */

dword
	rx_freq,
	tx_freq;

extern byte
	stack[],
	ram[];

extern word
	iv_sio[],
	iv_pio_a,
	iv_pio_b;

extern byte
	sio_b_rx_buf[],
	sio_b_rx_bufend[],
	sio_b_tx_buf[],
	sio_b_tx_bufend[],

	*sio_b_rx_ins,
	*sio_b_rx_rem,

	*sio_b_tx_ins,
	*sio_b_tx_rem,
	sio_b_tx_active;

extern void
	display(),
	key_setup(byte),
	quick_setup(byte);

byte
	(*cu_keypad)() = null_keypad;

void
	(*cu_update)()       = nop,
	(*cu_leds)()         = nop,
	(*cu_dtmf)(byte key) = nop,

	(*cu_volume)(); /* if can adjust volume */

struct buf
	*fx_tx_head, **fx_tx_tail = &fx_tx_head,
	*incoming_head, **incoming_tail = &incoming_head,
	*buf_free,
	buf_space[32];

byte keypad_XXX, led_XXX;

void
pio_b_enable()
{
	iow(PIO_B_CTRL, 0x83);
}

void
pio_b_disable()
{
	iow(PIO_B_CTRL, 0x03);
}

void
pll_strobe(byte which)
{
	iow(OUT_1, out_1 | which);
	iow(OUT_1, out_1);
}

static void
load_MC145156(byte strobe, word N, byte A)
{
	disable();

	pll_bits(0x02, 3);      /* SW1 and SW2 */

	pll_bits(0x02, N >> 8); /* 10 bits */
	pll_bits(0x80, N);

	pll_bits(0x40, A);      /* 7 bits */

	pll_strobe(strobe);
	restore();
}

/* MB87006 */

static void
load_pll(byte strobe, word R, word N, byte A)
{
	/* protected throughout, so that vco
	 * does not have so much time to
	 * wander off with new NA and old R.
	 * same reason for the order;
	 * 15 bits between strobes, not 18. */

	disable();

	pll_bits(0x02, N >> 8); /* 10 bits */
	pll_bits(0x80, N);

	pll_bits(0x40, A);      /* 7 bits */

	pll_bits(0x01, 0);      /* control bit '0' */
	pll_strobe(strobe);

	pll_bits(0x20, R >> 8); /* 14 bits */
	pll_bits(0x80, R);

	pll_bits(0x01, 1);      /* control bit '1' */
	pll_strobe(strobe);
	restore();
}

void
load_synth_ctrl()
{
#define SC_RX_LOW_BAND 0x01 /* 1 low band, 0 high */
#define SC_TX_LOW_BAND 0x02
#define SC_TX_BITS_1   0x04 /* prediv and pll, 0 on XXX keep supply here ? */
#define SC_TX_BITS_2   0x08 /* vco and amps, 1 on */

	byte x;

	if (tx_vco_on)
		x = SC_TX_BITS_2;
	else
		x = SC_TX_BITS_1;

#ifdef BASTARD
#define VBUFF 0x04

	if (tx_vco_on)
		x |= VBUFF;
	else
		x &= ~VBUFF;
#endif

	if (rx_freq < cfg.rx_band_center)
		x |= SC_RX_LOW_BAND;

	if (tx_freq < cfg.tx_band_center)
		x |= SC_TX_LOW_BAND;

	disable();
	pll_bits(0x08, cfg.deviation); /* XXX ls- or msbit first ? */
	pll_bits(0x08, x);

#ifdef PA_SCE
	pio_a_set(PA_SCE);       /* L8TM */
	pio_a_clr(PA_SCE);
#else
	pll_strobe(O1_SCE);      /* P8x */
#endif
	restore();
}

dword
vco_frequency(byte strobe, dword kHz)
{
	dword NA, N;
	word R;
	byte A;
	word z = 0;

	NA = kHz;
	switch (ulmoddiv(&NA, 25)) {
	case 24:
		z++;
	case 12:
		z++;
	case 0:
		R = TCXO / 12500;
		NA = NA * 2 + z;
		z = 50;
		break;

	case 18:
		z += 2;
	case 6:
		z++;
		R = TCXO / 6250;
		NA = NA * 4 + z;
		z = 25;
		break;

	default:
		R = TCXO / 10000;
		NA = kHz;
		if (ulmoddiv(&NA, 10) >= 5)
			NA++;
		z = 40;
	}

	N = NA;
	A = ulmoddiv(&N, cfg.prescaler);

	load_pll(strobe, R, N, A);

	if (N <= A || N > 0x3FF) /* XXX */
		kHz = 0;
	else
		kHz = NA * z / 4;

	return (kHz);
}

static byte *
RFC_of(dword kHz)
{
	/* one RFC value per every 512 kHz */

	return (&rfc_table[(kHz >> 9) & 0xFF]);
}

void
RFC_set(byte rfc)
{
	*RFC_of(rx_freq) = rfc;

	iow(DA_RFC, rfc);
}

byte
RFC_get()
{
	byte *r = RFC_of(rx_freq);

	while (*r == 0 && r > rfc_table)
		--r;

	return (*r);
}

/* TMR_LPF is MF-6 clock */

void
LPF(word hz)
{
	if (hz) {
		word count = (TMR_CLK / 100 + hz / 2) / hz;

		disable();
		iow(TMR_CTRL, TMR_SEL_LPF | TMR_BOTH | TMR_SQWAVE);
		iow(TMR_LPF, count);
		iow(TMR_LPF, count >> 8);
		restore();
	} else {
		/* mode 0, but count is not is written.
		 * output goes and stays LOW */

		iow(TMR_CTRL, TMR_SEL_LPF | TMR_LSB | TMR_INTTC);
	}
}

/* sidetone and ccir tx */

void
timer_mt(word hz)
{
	if (hz) {
		word count = (TMR_CLK + hz / 2) / hz;

		disable();
		iow(TMR_CTRL, TMR_SEL_MT | TMR_BOTH | TMR_SQWAVE);
		iow(TMR_MT, count);
		iow(TMR_MT, count >> 8);
		restore();
	} else {
		/* mode 0, but count is not is written.
		 * output goes and stays LOW */

		iow(TMR_CTRL, TMR_SEL_MT | TMR_LSB | TMR_INTTC);
	}
}

/* CLK2 originally 1968.75 Hz,
 * might have been rewired to 4.032 MHz (TMR_CLK)
 *
 * OUT2 must idle at '1' or FX429/OUT2/HOOK interrupts get blocked.
 *
 * currently unused. */

static void
timer_2(word hz)
{
#if 1 /* P8x thing */

	if (hz) {
		/* old ctcss tx was 64...254 (or 640...2540 if 4017) */

		word count = (TMR_CLK + hz / 2) / hz;

		disable();
		iow(TMR_CTRL, TMR_SEL_2 | TMR_BOTH | TMR_SQWAVE);
		iow(TMR_2, count);
		iow(TMR_2, count >> 8);
		restore();
	} else {
		/* mode 4, but count is not is written.
		 * output goes and stays HIGH */

		iow(TMR_CTRL, TMR_SEL_2 | TMR_LSB | TMR_SWTRIGG);
	}
#endif
}

static void
timer_brg()
{
#if 0 /* L8x thing */

	/* XXX L8x could provide other baudrates than 2400,4800,9600 */

	word hz = 16 * 9600;
	word count = (TMR_CLK + hz / 2) / hz;

	disable();
	iow(TMR_CTRL, TMR_SEL_BRG | TMR_BOTH | TMR_SQWAVE);
	iow(TMR_BRG, count);
	iow(TMR_BRG, count >> 8);
	restore();
#endif
}

static const word ccir_tones[] = {
	1981, 1124, 1197, 1275, 1358,
	1446, 1540, 1640, 1747, 1860,
	2400, /* A */
	 930, /* B */
	2247, /* C */
	 991, /* D */
	2110, /* E */
	   0, /* F */
};

void
ccir_send(byte *s)
{
	word hz, len = 100;
	byte x, ox = -1;

	timer_mt(0);
	out_0_set(O0_CCIRC);

	while (x = *s++) {
		if (x == 'L') {
			len = 500;         /* L changes next duration to 500 msec */
			continue;
		}
		x = nybble(x);
		if (x & ~0xF)
			x = 0xF;           /* other non-hex makes a pause */

		if (x == ox)
			x = 0xE;
		ox = x;
		hz = ccir_tones[x];
		LPF(hz);
		timer_mt(hz);

		delay_msec(len);
		len = 100;
	}
	out_0_clr(O0_CCIRC);
	timer_mt(0);

	LPF(cfg.lpf);
}

void
dtmf(byte c)
{
	(*cu_dtmf)(c);
}

void
leds()
{
	led_XXX = 0;

	(*cu_leds)();
}

void
lit(byte *led)
{
	if (*led == 0) {
		*led = 1;
		led_XXX = 1;
	}
}

void
dim(byte *led)
{
	if (*led) {
		*led = 0;
		led_XXX = 1;
	}
}

/* optional ctcss tx --- RFC---MOD wiring needed */

void
ctcss(word dHz)
{
	extern byte ctcss_sine[];
	static byte old_gain;
	byte center = RFC_get(); /* XXX 0x80 if not cfg.fdx */

	if (center != ctcss_sine[0] || cfg.ctcss_gain != old_gain) {
		old_gain = cfg.ctcss_gain;
		make_sine(ctcss_sine, cfg.ctcss_gain, center);
	}

	/* XXX * 0x10000U * 2048 / TMR_CLK / 10 */

	ctcss_go(dHz * 0x40000U / 78750);
}

static void
close_audio()
{
	out_0_clr(O0_AUDIOC);
}

static void
open_audio()
{
	if (cfg.volume)
		out_0_set(O0_AUDIOC);
}

static void
close_squelch()
{
	if (!audio_flags)
		close_audio();

	squelch_timer = 0;
	squelch_open = 0;

	dim(&led_serv);

	pio_b_input(PB_EXIN1); /* pullup */
}

static void
open_squelch()
{
	if (!audio_flags)
		open_audio();

	squelch_timer = 0;
	squelch_open = 1;

	lit(&led_serv);

	pio_b_output(PB_EXIN1); /* output 0 */
}

void
receive(dword kHz)
{
	rx_freq = vco_frequency(O1_SRE, kHz + cfg.if_offset)
	        - cfg.if_offset;

	load_synth_ctrl();     /* rx a/b vco-band */

	iow(DA_RFC, RFC_get());
}

byte
legal_tx(dword kHz)
{
	if (kHz > cfg.tx_band_start
	 && kHz < cfg.tx_band_end) {

		if (kHz >  420000 && kHz <  450000) return (1);
		if (kHz >  144000 && kHz <  148000) return (1);
		if (kHz >   50000 && kHz <   54000) return (1);
		if (kHz > 1240000 && kHz < 1300000) return (1);
		if (kHz >  902000 && kHz <  928000) return (1);
		if (kHz >  220000 && kHz <  225000) return (1);
		if (kHz >   70000 && kHz <   70500) return (1);
	}
	return (0);
}

void
transmit_off()
{
	word mark;

	encoder_turnoff();      /* XXX just testing */

	iow(DA_TPC, 0);

	mark = timef_mark(8 MSEC);

	LPF(0);
	dtmf(0);
	encoder_end();

	dim(&led_roam);
	leds();
	display();

	while (! timef_passed(mark))
		;

	out_1_change(O1_TXON | O1_TXOFF, O1_TXOFF);

	tx_vco_on = 0;
	load_synth_ctrl();

	audio_flags &= ~0x80;  /* sql_open follows */

	if (audio_flags & 3)   /* monitor pressed or toggled on */
		open_audio();

	iow(DA_RFC, RFC_get()); /* might have been fouled */

	transmitting = 0;
}

void
transmit_on()
{
	word mark;

	transmitting = 1;

	iow(DA_TPC, 0);

	tx_freq = cfg.rx_freq + cfg.tx_offset;

	if (!cfg.fdx) {
		audio_flags |= 0x80; /* half-duplex tx disables squelch */
		close_audio();
		close_squelch();
	}

	out_1_change(O1_TXON | O1_TXOFF, O1_TXOFF);

	tx_vco_on = 1;
	load_synth_ctrl(); /* tx a/b vco-band, supplies, deviation */

	tx_freq = vco_frequency(O1_STE, tx_freq);

	/* overlap the delay with useful stuff */

	mark = timef_mark(25 MSEC);

	/* power only if amateur band */

	if (legal_tx(tx_freq)) {

		LPF(cfg.lpf);      /* tx audio lowpass */

		lit(&led_roam);
		display();
		leds();

		if (cfg.ctcss_gain && cfg.ctcss)
#if 0
			dcs(cfg.ctcss);
#else
			ctcss(cfg.ctcss);
#endif

		out_1_change(O1_TXON | O1_TXOFF, O1_TXON);

		while (! timef_passed(mark))
			;

		iow(DA_TPC, cfg.tpc);
	}
}

void
lights_on()
{
	if (cfg.lights)
		light_timo = time_mark(cfg.lights);

	lit(&led_lcdlight);
	lit(&led_keylight);
}

void
lights_off()
{
	light_timo = 0;

	dim(&led_lcdlight);
	dim(&led_keylight);
}

static void
volume(int add)
{
	byte x = 0;

	/* silent service
	 * ear only
	 * 8 speaker levels */

	if (add > 0)
		cfg.volume++;

	if (add < 0 && cfg.volume)
		cfg.volume--;

	if (cfg.volume > 9)
		cfg.volume = 9;

	if (cu_volume) {
		(*cu_volume)();

		x |= O0_INH;
	} else {
		if (cfg.volume >= 2)
			x |= (cfg.volume - 2) & O0_VOLUME; /* XXX rightjusted */
		else
			x |= O0_INH;
	}
	out_0_change(O0_VOLUME | O0_INH, x);

	disable();
	if (cfg.volume == 0)
		close_audio();
	else
	if (squelch_open || (audio_flags & 3))
		open_audio();
	restore();
}

extern byte cu58af_buttons;

byte
ptt_pressed()
{
	return ((sio_b_rr0 & SB_PTT) | (cu58af_buttons & 0x40));
}

static void
qsy()
{
	disable();
	close_squelch();
	restore();

	tx_freq = cfg.rx_freq + cfg.tx_offset;

	receive(cfg.rx_freq);
}

static void
frequency(dword kHz)
{
	cfg.rx_freq   = kHz;
	cfg.tx_offset = 0;        /* XXX duplex band(s) */
	cfg.ctcss     = 0;
	qsy();
}

static void
memory(word n)
{
	struct memory *p = &memories[n];

	if (p->rx_freq) {
		cfg.rx_freq   = p->rx_freq;
		cfg.tx_offset = p->tx_offset;
		cfg.ctcss     = p->ctcss;
		qsy();
		last_memory = n;
	}
}

static void
key_make(byte c)
{
	if (ptt_pressed()) {
		dtmf(c); /* XXX latch timeout and break actions */
		return;
	}
	if (in_setup && c != 'B') {
		key_setup(c);
		return;
	}
	if (isdigit(c)) {
		if (keycnt < display_2_len) {
			buffer[keycnt++] = c;
			bzero(buffer + keycnt, sizeof(buffer) - keycnt); /* XXX */
		}
		return;
	}
	switch (c) {
	case 'C':
		if (keycnt)
			buffer[--keycnt] = 0;
		break;

	case 'E':
		keycnt = 0;
		in_setup = 1;
		break;

	case '#':
		switch (keycnt) {
		case 0:
			break;

		case 1:
		case 2:
			memory(atow(buffer));
			break;

		case 3:
			frequency(atodw(buffer) + cfg.implied * 1000UL);
			break;
		case 4:
			frequency(atodw(buffer) + cfg.implied / 10 * 10000UL);
			break;
		default:
			frequency(atodw(buffer));
		}
		keycnt = 0;
		break;

	case 'S':

		timer_mt(0);
		out_0_set(O0_MTC);
		silly();
		out_0_clr(O0_MTC);
		timer_mt(0);
		break;

	case 'R':
	case '*':
		keycnt = 0;
		break;

	case '+':
		volume(+1);
		break;

	case '-':
		volume(-1);
		break;

	case 'B':
		audio_flags |= 2; /* monitor while B pressed */
		open_audio();
		if (cfg.tx_offset)
			receive(cfg.rx_freq + cfg.tx_offset);
		break;
	}
}

static void
key_break(byte c)
{
	if (ptt_pressed()) {
		dtmf(0);
		return;
	}
	switch (c) {
	case 'B':
		if (cfg.tx_offset)
			receive(cfg.rx_freq);

		disable();
		audio_flags &= ~2;
		if (key_timo)
			audio_flags ^= 1; /* quick B, toggle */

		if (!audio_flags && !squelch_open)
			close_audio();

		restore();

		icon_star = (audio_flags & 1);
		display();
		break;
	}
	in_setup &= ~2; /* display returns to normal later */
}

/* if the key had been released before timeout,
 * 'key' is 0 but last_key still something */

static void
key_timeout()
{
	byte d = 0;

	key_timo = 0;

	if (ptt_pressed()) { /* XXX re-enable mic ? */
		return;
	}
	if (alphamode) {
		alphamode = 0;
		d = 1;
	}

	/* long keypresses */

	switch (key) {
	case 'C':
		if (keycnt) {
			keycnt = 0;
			d = 1;
		}
		break;

	case '-':
		if (!in_setup && cfg.volume == 1) /* no autorepeat to mute */
			break;
	case '+':
		key_make(key);
		d = 1;
		key_timo = timef_mark(200 MSEC); /* from now, break or no */
		break;

	/* XXX */

	case '1':
	case '4':
		keycnt = 0;
		in_setup |= 2;
		quick_setup(key);
		d = 1;
		key_timo = timef_mark(200 MSEC); /* from now, break or no */
		break;

	case '2':
	case '5':
		{
			byte n = last_memory;

			do {
				if (key == '2') {
					if (++n >= 100)
						n = 0;
				} else {
					if (--n >= 100)
						n = 99;
				}
				if (memories[n].rx_freq) {
					memory(n);
					break;
				}
			} while (n != last_memory);
		}
		keycnt = 0;
		d = 1;
		key_timo = timef_mark(750 MSEC); /* from now, break or no */
		break;

	case '3':
		rx_freq += 50;
	case '6':
		rx_freq -= 25;
		frequency(rx_freq);
		keycnt = 0;
		d = 1;
		key_timo = timef_mark(750 MSEC); /* from now, break or no */
		break;
	}

	if (d)
		display();
}

static void
keypad()
{
	byte c;

	keypad_XXX = 0;

	c = (*cu_keypad)();
	if (c == 0xFF || c == key)  /* cannot tell or false alarm */
		return;

	if (key && c)
		key_break(key); /* chord */

	key = c;

	lights_on();         /* updated later XXX long synchronous action ? */

	if (c) {
		key_make(c);
		last_key = c;
		key_timo = timef_mark(750 MSEC); /* from now, break or no */
		display();
	} else {
		key_break(last_key);
	}
}

static void
progress(byte c)                         /* printf debugger */
{
	extern void cu53_update();

	display_2[0] = c;

	if (cu_update != nop)
		(*cu_update)();
	else
		cu53_update();
}

/* matches table in setup.c */

static const byte
	sio_formats[] = {
		WR4_X64_CLK, /* 2400 */
		WR4_X32_CLK, /* 4800 */
		WR4_X16_CLK, /* 9600 */

		WR4_X64_CLK | WR4_PARITY_ENB | WR4_PARITY_EVEN,
		WR4_X32_CLK | WR4_PARITY_ENB | WR4_PARITY_EVEN,
		WR4_X16_CLK | WR4_PARITY_ENB | WR4_PARITY_EVEN,

		WR4_X64_CLK | WR4_PARITY_ENB | WR4_PARITY_ODD,
		WR4_X32_CLK | WR4_PARITY_ENB | WR4_PARITY_ODD,
		WR4_X16_CLK | WR4_PARITY_ENB | WR4_PARITY_ODD,
	};

static byte
sio_format(byte i)
{
	if (i >= sizeof(sio_formats))
		i = 0;

	return (sio_formats[i]);
}

/* in place of the pibolar brom, chipselect /CSPROM.
 * A3 might work as /WR. or might not work. */

static void
init_MT8888()
{
	ior(MT8888_RS);
	iow(MT8888_WC, 0x0);
	iow(MT8888_WC, 0x0);

	iow(MT8888_WC, 0x8); /* CRA: RSEL 1  IRQ  0  CP/DTMF# 0  TOUT   0 */
	iow(MT8888_WC, 0x1); /* CRB: C/R# 0  S/D# 0  TEST     0  BURST# 1 */

	ior(MT8888_RS);
}

/* interrupts disabled, will be enabled after this returns */

void
initialise()
{
	int i;

	iow(WDOG, 0);
	iow(DA_TPC, 0);

	/* sometimes battery a/d sticks at FF. input > vcc latchup ?
	 * maybe this unlatches, mux kept at IN7 == GND for a longish time.
	 * no, it doesn't help... XXX unsolvable ? */

	iow(AD_IN7, 0);

	/* select 48 kB contiguous ROM0
	 * and 16 kB work ram.
	 * O2_SMEM is really ROMA16 on P8E, 27C512 not affected.
	 * bootstrap has done something like this already. */

	iow(CSMEM, 1);        /* P8E SMEM */

	iow(OUT_0, out_0);
	iow(OUT_1, out_1);
	iow(OUT_2, out_2);    /* ROM0 and P8N SMEM */

	/* lml */

	timer_mt(444);
	out_0_change(O0_MTC | O0_VOLUME, O0_MTC | 2);

	timer_2(0);        /* used to be optional ctcss tx */
	timer_brg();
	iow(CSKEY, 0);     /* P8E bipolar prom (not fitted) +5v on */

	iow(FX429_CSR, 4); /* modem reset */
	delay_usec(3000);  /* delay_msec would require ticking */
	iow(FX429_CSR, 0);

	/* PIO-A */

	iow(PIO_A_DATA, pio_a);        /* all inputs though                 */
	iow(PIO_A_CTRL, 0);            /* in case it was waiting something  */

	iow(PIO_A_CTRL, (byte) &iv_pio_a);
	iow(PIO_A_CTRL, 0xCF);         /* operating mode 3                  */
	iow(PIO_A_CTRL, pio_a_ddr);    /* 1 input, 0 output pin             */
	iow(PIO_A_CTRL, 0x97);         /* interrupt control: IE OR LOW MASK */
	iow(PIO_A_CTRL, 0xFE);         /* monitor mask, just PA.0           */

	/* PIO-B */

	iow(PIO_B_DATA, pio_b);
	iow(PIO_B_CTRL, 0);

	iow(PIO_B_CTRL, (byte) &iv_pio_b);
	iow(PIO_B_CTRL, 0xCF);
	iow(PIO_B_CTRL, pio_b_ddr);
	iow(PIO_B_CTRL, 0x97);
	iow(PIO_B_CTRL, 0xFF);         /* monitor mask, nothing */

	/* cfg
	 * after PIO init. PIOB.0 == RAMA12 must be valid with P8N */

	display_2[7] = hexdig[logic];
	progress('0');

	load_bbram();
	new_defaults();

	progress('1');

	/* SIO-A */

	ior(SIO_A_CSR);
	iow(SIO_A_CSR, WR0_CHANNEL_RESET);
	nop();

	iow(SIO_A_CSR, 4);
	iow(SIO_A_CSR, WR4_1STOPBIT | sio_format(cfg.sio_a_fmt));
	iow(SIO_A_CSR, 3);
	iow(SIO_A_CSR, WR3_8BIT_RX | WR3_RX_ENB);
	iow(SIO_A_CSR, 5);
	iow(SIO_A_CSR, sio_a_wr5);
	iow(SIO_A_CSR, 1);
	iow(SIO_A_CSR, WR1_STATUS_VECTOR
	             | WR1_ESCINT_ENB | WR1_TXINT_ENB | WR1_INT_ALLC);

	ior(SIO_A_CSR);
	ior(SIO_A_DATA);
	iow(SIO_A_CSR, WR0_ERROR_RESET);
	iow(SIO_A_CSR, WR0_RESET_ESCINT);

	sio_a_rr0 = ior(SIO_A_CSR);

	/* SIO-B */

	ior(SIO_B_CSR);
	iow(SIO_B_CSR, WR0_CHANNEL_RESET);
	nop();

	iow(SIO_B_CSR, 4);
	iow(SIO_B_CSR, WR4_1STOPBIT | sio_format(cfg.sio_b_fmt));
	iow(SIO_B_CSR, 3);
	iow(SIO_B_CSR, WR3_8BIT_RX | WR3_RX_ENB);
	iow(SIO_B_CSR, 5);
	iow(SIO_B_CSR, sio_b_wr5);
	iow(SIO_B_CSR, 1);
	iow(SIO_B_CSR, WR1_STATUS_VECTOR
	             | WR1_ESCINT_ENB | WR1_TXINT_ENB | WR1_INT_ALLC);

	iow(SIO_B_CSR, 2);
	iow(SIO_B_CSR, (byte) iv_sio);

	ior(SIO_B_CSR);
	ior(SIO_B_DATA);
	iow(SIO_B_CSR, WR0_ERROR_RESET);
	iow(SIO_B_CSR, WR0_RESET_ESCINT);

	sio_b_rr0 = ior(SIO_B_CSR);

	progress('2');

	/* handset, if any */

	cu_probe();

	/* XXX */

	lights_on();
	leds();

	progress('3');

	sio_a_puts(version); /* XXX */
	sio_a_puts("\r\n");

	/* enable receive */

	qsy();

	/* all packets to free list */

	for (i = 0; i < __SIZE(buf_space); ++i)
		buf_relse(&buf_space[i]);

	fx429_cmd = FX429_RXENB;
	iow(FX429_CSR, fx429_cmd);

	init_MT8888();

	/* reset HOOK edge detective (and TMR_2_OUT latch) */

	ior(TMR_CTRL);

	/* stop lml */

	timer_mt(0);
	out_0_clr(O0_MTC);

	volume(+0);

	progress('-');

	batt_timo = time_mark(60);
}

/* XXX */

struct buf *
buf_alloc()
{
	struct buf *p;

	disable();

	p = buf_free;
	if (p)
		buf_free = p->next;

	restore();

	return (p);
}

void
buf_relse(struct buf *p)
{
	disable();

	p->next = buf_free;
	buf_free = p;
	restore();
}

byte
fx_check(byte *p, word datalen)
{
	word crc = crc_ccitt(p, datalen);

	return (p[datalen + 0] == (crc >> 8)     /* MSByte 1st */
	     && p[datalen + 1] == (crc & 0xFF));
}

void
fx429_send(byte *pkt, word datalen)
{
	struct buf *p = buf_alloc();
	word crc;

	if (!p)
		boot();

	p->next = NULL;
	p->type = FSK;
	p->len = 3 + datalen + 2;

	p->data[0] = 0xAA; /* 8 idlebits between packets, except ... */
	p->data[1] = 0xC4; /* SYNC */
	p->data[2] = 0xD7;

	bcopy(pkt, p->data + 3, datalen);

	crc = crc_ccitt(pkt, datalen);
	p->data[3 + datalen] = crc >> 8;     /* MSByte 1st */
	p->data[4 + datalen] = crc;

	disable();

	*fx_tx_tail = p;
	fx_tx_tail = &p->next;

	if ((fx429_cmd & FX429_TXENB) == 0) {
		fx429_cmd |= FX429_TXENB;
		iow(FX429_CSR, fx429_cmd);

		iow(FX429_DATA, 0xAA); /* ... 16 additional idlebits */
		iow(FX429_DATA, 0xAA); /* fills holding register */
	}
	restore();
}

void
ax_send(byte *pkt, word len)
{
	word idle, crc;

	/* abort-tone */

	timer_mt(1200);
	out_0_set(O0_CCIRC);

	/* pwm slower than 15.750 kHz (256 timer clocks),
	 * full range in sine ok */
	{
		extern byte afsk_sine[];
		static byte old_gain;

		if (cfg.afsk_gain != old_gain) {
			old_gain = cfg.afsk_gain;
			make_sine(afsk_sine, cfg.afsk_gain, 0x80);
		}
	}
	crc = crc_ccitt(pkt, len);
	pkt[len++] = crc;
	pkt[len++] = crc >> 8;

	/* msec to doublebits, approx */

	idle = cfg.ax_txdelay / 2;
	idle += idle / 4;

	axout(pkt, len, idle);

	out_0_clr(O0_CCIRC);
	timer_mt(0);
}

/*
 *  hare/tortoise/slug is called repeatedly, for ever.
 *  between calls, boot.s does
 *    enable()
 *    fido = 0
 *
 *  must return in 10 seconds, or reboot.
 */

static void
ax_test()
{
	byte buf[100];

	axaddr(buf + 0, "QST");
	axaddr(buf + 7, cfg.mycall);
	buf[13] |= 0x01;
	buf[14] = 0x03;
	buf[15] = 0xF0;
	strcpyx(buf + 16, "hoplofobinen hopliitti");

	ax_send(buf, 16 + 22);
}

static void
fx_test()
{
	fx429_send("\xC1\x61\x60\x12\x34\x5F", 6);
	fx429_send("\xC1\x61\x60\x22\x33\x55", 6);
}

/* XXX */

static void
handle_incoming(struct buf *p)
{
	int i;

	switch (p->type) {

	case FSK:
		p->len -= 2;

		if (! fx_check(p->data, p->len))
			break;

		sio_a_puts("FSK  ");
		for (i = 0; i < p->len; ++i) {
			sio_a_putc(hexdig[p->data[i] >> 4]);
			sio_a_putc(hexdig[p->data[i] & 15]);
		}
		sio_a_putc('\n');
		break;

	case CCIR:
		sio_a_puts("CCIR ");
		for (i = 0; i < p->len; ++i) {
			sio_a_putc(p->data[i]);
		}
		sio_a_putc('\n');
		break;

	case DTMF:
		sio_a_puts("DTMF ");
		for (i = 0; i < p->len; ++i) {
			sio_a_putc(p->data[i]);
		}
		sio_a_putc('\n');
		break;
	}
}

/* as fast as possible */

void
hare()
{
	word c;

#if 0
	/* "I'm here" */

	out_2_toggle(O2_ROMA15);
#endif

#define gotc(c) (((c) & 0xFF00) == 0) /* cc likes it */

	c = sio_a_getc();
	if (gotc(c))
		nmea_input(c);

	if (ptt_pressed()) {
		if (! transmitting && ptt_works) {
			transmit_on();

			if (keycnt == 1)       /* XXX */
				ax_test();

			if (keycnt == 2)
				fx_test();

			if (keycnt >= 3)
				ccir_send(buffer);

			out_0_clr(O0_MICM);
		}
	} else {
		ptt_works = 1;
		if (transmitting && (fx429_cmd & FX429_TXENB) == 0) {
			out_0_set(O0_MICM);
			transmit_off();
		}
	}
}

/* every 4 msec or so */

void
tortoise()
{
	static byte
		pa,
		pb = PB_EXIN2 | PB_TMR0;

	byte diff;

	if (keypad_XXX)
		keypad();

	if (led_XXX)
		leds();

	if (key_timo && timef_passed(key_timo))
		key_timeout();

	if (light_timo && time_passed(light_timo))
		lights_off();

	diff = pa;
	pa = ior(PIO_A_DATA);
	diff ^= pa;

	if (diff & PA_OFFHOOK) {
		lights_on();
	}

	diff = pb;
	pb = ior(PIO_B_DATA);
	diff ^= pb;

	if (diff & PB_EXIN2) {
		sio_a_puts("EXIN2\n");
	}
	if (diff & PB_TMR0) {
		sio_a_puts("TMR_2_OUT\n");
	}

	if (incoming_head) {
		struct buf *p;

		disable();

		if (p = incoming_head) {
			if ((incoming_head = p->next) == NULL)
				incoming_tail = &incoming_head;
		}
		restore();

		if (p) {
			handle_incoming(p);
			buf_relse(p);
		}
	}
}

/* every second or so */

void
slug()
{
	if (millivolt(ad_batt) < cfg.batt_cutoff) {
		if (time_passed(batt_timo))
			power_off();

		if (led_on)
			dim(&led_on);
		else
			lit(&led_on);
	} else {
		batt_timo = time_mark(60);
		lit(&led_on);
	}

	/* adc latched up ? or what it is ... */

	if (ad_batt == 0xFF)
		lit(&led_call);
	else
		dim(&led_call);

	display(); /* XXX */
}

void
power_off()
{
	disable();

	iow(DA_TPC, 0);

	if (cfg.magic == MAGIC)
		save_bbram();

	iow(PIO_B_DATA, pio_b | PB_PWROFF);

	delay_usec(1000);
	boot();             /* d/c to Abiko ? :) */
}

/*
 *  NMI causes
 *  - RESET pulse from watchdog (doesn't create /RESET)
 *    or loss of supply (cpu /RESET in 1 msec or so)
 *  - on/off button changes to off-position (bounces sometimes)
 *
 *  return is safe, if not much is done.
 *
 *  if it is/was 'RESET', output ports are/were undefined XXX arg-g-h...
 */

/* power off by hw in 1 second, if these get stuck */

static void
dump_drain()
{
	ior(SIO_A_CSR); /* RR0, please */

	do {
		iow(SIO_A_CSR, 1);
	} while ((ior(SIO_A_CSR) & RR1_ALL_SENT) == 0);
}

static void
dump_char(byte b)
{
	ior(SIO_A_CSR); /* RR0, please */

	while ((ior(SIO_A_CSR) & RR0_TBE) == 0)
		;

	iow(SIO_A_DATA, b);
}

static void
dump_str(byte *s)
{
	while (*s)
		dump_char(*s++);
}

static void
dump_byte(byte b)
{
	dump_char(hexdig[b >> 4]);
	dump_char(hexdig[b & 15]);
}

static void
dump_word(word w)
{
	dump_byte(w >> 8);
	dump_byte(w);
	dump_str("\n");
}

/* active stack contents */

static void
crash()
{
	word here, *sp = &here + 1;

	dump_str("\nugh...\n");

	dump_word((word) sp);

	if (sp >= (word *) ram) {
		do {
			dump_word(*sp);
		} while (++sp < (word *) stack);
	}
	/* or else ? */

	dump_drain();
	boot();
}

void
nmi()                     /* ra1 iy ix de' bc' af' hl' de bc af hl ra2 ... */
{
	byte why = ior(PIO_A_DATA);

	if (why & PA_WDR)
		crash();          /* power off by hw in 1 second */

	if (why & PA_PWR) {   /* switch off-position is '1' */
		if (cfg.stay_on)
			return;       /* ignore on/off */
		power_off();
	}
	/* loss of supply then XXX */

	iow(DA_TPC, 0);

	if (cfg.magic == MAGIC)
		save_bbram();
}

void
incoming(struct buf *p, byte type)
{
	p->next = NULL;
	p->type = type;

	disable();
	*incoming_tail = p;
	incoming_tail = &p->next;
	restore();
}

/* XXX decoders could collect directly to a buf
 * XXX decoders should rewind on qsy XXX */

void
incoming_b(byte *buf, byte len, byte type)
{
	struct buf *p = buf_alloc();

	if (p) {
		bcopy(buf, p->data, p->len = len);
		incoming(p, type);
	}
}

static byte ccir_n;

void
ccir_decode(byte c)
{
	static word timo1, timo2;
	static byte buf[15];

	if (c == 0xF) {

		/* ignore if truncated or too short or long duration.
		 * XXX "123" would be 3 * 100 msec
		 * XXX leading tone might be up to 1 second long ? */

		if (ccir_n && ccir_n < sizeof(buf)
		 && timef_passed(timo1) && ! time_passed(timo2))
			incoming_b(buf, ccir_n, CCIR);

		ccir_n = 0;
	} else {
		if (ccir_n < sizeof(buf)) {
			if (ccir_n && c == 0xE)
				c = buf[ccir_n - 1];    /* repeat */
			else
				c = hexdig[c];

			buf[ccir_n++] = c;
			if (ccir_n == 1) {
				timo1 = timef_mark(300 MSEC);
				timo2 = time_mark(5);
			}
		}
	}
}

/* MT8870, MT8888, MT8889 databits are all equal.
 * bit 0x10 is 8870-like 'StD', emulated if not real 8870.
 *
 * XXX MT8889 timing is incompatible with Z80,
 * XXX A1 must be used as R/#W.
 * XXX gaaah, seems not work. and 8888 timing looks unsuitable too */

static const byte dtmf_8870_tab[] = "D1234567890*#ABC";

static byte dtmf_n, dtmf_buf[15];

void
dtmf_timeout()
{
	/* ignore if truncated */

	if (dtmf_n && dtmf_n < sizeof(dtmf_buf))
		incoming_b(dtmf_buf, dtmf_n, DTMF);

	dtmf_n = 0;
	dtmf_timo = 0;
}

void
dtmf_decode(byte c)
{
	if (c & 0x10) {
		if (dtmf_n < sizeof(dtmf_buf))
			dtmf_buf[dtmf_n++] = dtmf_8870_tab[c & 0xF];

		dtmf_timo = 0;

	} else if (dtmf_n)
		dtmf_timo = timef_mark(5000 MSEC); /* 5 seconds after StD drops */

}

/*
 *  pio_a_int (1968.75 Hz, highest priority)
 *  calls tick() at 256 Hz,
 *  skipping if previous tick has not finished yet.
 *
 *  timex has been incremented,
 *  PIO acked and interrupts enabled
 *
 *  must execute every 0.5 seconds or reboot.
 */

void
tick()
{
	iow(WDOG, 0);

	if (++fido >= 10 * TICK_HZ)
		crash();

	if (cfg.stay_on)
		sio_b_toggle(SB_MON);

	/* if FX003 is not there,
	 * pulled up pio pins show 0xF = notone */

	{
		static byte old = 0xF;
		byte c = read_ccir();

		if (c != old) {
			old = c;
			ccir_decode(c);
		}
	}
	/* read_dtmf returns StD D3 D2 D1 D0
	 * if there is no decoder, bus can show changing garbage.
	 * parameter, no autodetect */

	if (cfg.dtmf_decoder) {
		static byte old;
		byte c = read_dtmf(cfg.dtmf_decoder);

		if (c != old) {
			old = c;
			dtmf_decode(c); /* StD in 0x10 */
		} else {
			if (dtmf_timo && timef_passed(dtmf_timo))
				dtmf_timeout();
		}
	}
	/* squelch still runs even if monitor is on,
	 * though audio path is not touched */

	if ((audio_flags & 0x80) == 0) {
		byte ad = cfg.sql_method ? ad_rssi : ad_sql;

		if (squelch_open) {
			if (ad >= cfg.sql_level)
				squelch_timer = 0;
			else
			if (ad_rssi_25ms_ago > cfg.sql_bi_rssi
			 || ++squelch_timer >= cfg.sql_close / 4) /* XXX */
				close_squelch();
		} else {
			if (ad < cfg.sql_level + cfg.sql_hyst)
				squelch_timer = 0;
			else
			if (++squelch_timer >= cfg.sql_open / 4) /* XXX */
				open_squelch();
		}
	}
}

struct buf *in_pkt;
byte in_pkt_ix, in_pkt_len;

/* sometimes a false alarm, caused by HOOK or TMR_2 */

static void
fx429_interrupt()
{
	byte c, stat;

	/* XXX extra dancing around edge sensitivity */

	stat = ior(FX429_CSR);      /* clears /INT unless RXRDY */

	if (stat & FX429_RXRDY) {
		byte nstat;             /* stat |= ior() would confuse cc */

		c = ior(FX429_DATA);    /* clears /INT from RXRDY */

		nstat = ior(FX429_CSR); /* clears new /INT from other sources */
		stat |= nstat;
	}
	if (stat & FX429_TXRDY) {
		static byte out_pkt_ix;

		struct buf *p = fx_tx_head;

		if (p) {
			if (stat & FX429_TXIDL) /* oopsidaisies, try again */
				out_pkt_ix = 0;

			iow(FX429_DATA, p->data[out_pkt_ix++]);

			if (out_pkt_ix == p->len) {
				out_pkt_ix = 0;
				if ((fx_tx_head = p->next) == NULL)
					fx_tx_tail = &fx_tx_head;
				buf_relse(p);
			}
		} else {
			if (stat & FX429_TXIDL) {
				fx429_cmd &= ~FX429_TXENB;
				iow(FX429_CSR, fx429_cmd);
			}
		}
	}
	if (stat & FX429_SYNC) {
		in_pkt_ix = 0;
		if (in_pkt == NULL)
			in_pkt = buf_alloc();
	}
	if (stat & FX429_RXRDY) {
		if (in_pkt)
			in_pkt->data[in_pkt_ix] = c;

		if (++in_pkt_ix == 1) {
			if ((c & 0xF0) == 0xC0)
				in_pkt_len = 8;
			else
				in_pkt_len = 15;

			fx429_cmd |= FX429_RXFMT;         /* crc by software */
			iow(FX429_CSR, fx429_cmd);

		} else if (in_pkt_ix == in_pkt_len) {
			if (in_pkt) {
				in_pkt->len = in_pkt_len;
				incoming(in_pkt, FSK);
			}
			in_pkt_ix = 0;
			in_pkt = NULL;

			fx429_cmd &= ~FX429_RXFMT;    /* wait sync again */
			iow(FX429_CSR, fx429_cmd);
		}
	}
}

/*
 *  sio_a_rca, sio_a_tbe, sio_a_src,
 *  sio_a_getc, sio_a_putc,
 *  in boot.s
 */

/*
 *  FX429 or HOOK or TMR_2
 *  CU53 DA or CU58AF /INT
 *  SYNC_A
 *  BREAK character detected or terminated
 */

void
sio_a_esc()
{
	byte diff;

	/* reset HOOK edge detective (and TMR_2_OUT latch) */

	ior(TMR_CTRL);

	diff = sio_a_rr0;
	sio_a_rr0 = ior(SIO_A_CSR);
	diff ^= sio_a_rr0;

	/* falling edge of FX429 /INT or TMR_2_OUT
	 * or either edge of HOOK
	 * toggles 74HC107 */

	if (diff & SA_KKINT)
		fx429_interrupt();

	/* change of handset DA aka /INT */

	if (diff & SA_DA)
		keypad_XXX = 1;

	iow(SIO_A_CSR, WR0_RESET_ESCINT);
}

/*
 *  sio_b_xxx
 *  good candidates to boot.s
 *  or just ++(byte)p
 */

#define sio_b_tx_bump(p) \
	do { if (++p == sio_b_tx_bufend) p = sio_b_tx_buf; } while (0)

#define sio_b_rx_bump(p) \
	do { if (++p == sio_b_rx_bufend) p = sio_b_rx_buf; } while (0)

void
sio_b_putc(word c)
{
	disable();

	if (sio_b_tx_active || (ior(SIO_B_CSR) & SB_DMIDLE) == 0) {
		*sio_b_tx_ins = c;
		sio_b_tx_bump(sio_b_tx_ins);
	} else {
		iow(SIO_B_DATA, c);
		sio_b_tx_active = 1;
	}
	restore();
}

void
sio_b_start()
{
	iow(SIO_B_DATA, *sio_b_tx_rem);
	sio_b_tx_bump(sio_b_tx_rem);
	sio_b_tx_active = 1;
}

void
sio_b_rca()
{
	*sio_b_rx_ins = ior(SIO_B_DATA);
	sio_b_rx_bump(sio_b_rx_ins);
}

void
sio_b_tbe()
{
	if (sio_b_tx_rem != sio_b_tx_ins) {
		iow(SIO_B_DATA, *sio_b_tx_rem);
		sio_b_tx_bump(sio_b_tx_rem);
	} else {
		iow(SIO_B_CSR, WR0_RESET_TXINT);
		sio_b_tx_active = 0;
	}
}

void
sio_b_src()
{
	/* parity, framing or overrun. just discard */

	ior(SIO_B_DATA);
	iow(SIO_B_CSR, WR0_ERROR_RESET);

	/* XXX if sending, rewind, then start or defer */
}

/*
 *  /PTT
 *  /LOCAL
 *  DMIDLE (aka NETFREE)
 *  BREAK character detected or terminated
 *
 *  lowest priority
 */

void
sio_b_esc()
{
	sio_b_rr0 = ior(SIO_B_CSR);

	iow(SIO_B_CSR, WR0_RESET_ESCINT);

	if ((sio_b_rr0 & SB_DMIDLE)
	 && sio_b_tx_active == 0
	 && sio_b_tx_rem != sio_b_tx_ins)
		sio_b_start();
}

/*
 *  monitormask is currently empty
 *
 *  could be used to catch EXIN1/2
 *  or another periodic (from TMR_2, rewiring)
 *  or sw uart input to DCU
 */

void
pio_b_int()
{
}

