/*
 *  cold reset from 0xFFFF:0 (last 16 bytes of rom)
 *  fixes UMCS for 256 kB rom and
 *  jumps to here, boot: in 0xC000:0x0100
 */

#include "head.h"
#include "pcb.h"
#include "io.h"

#asm

HZ equ 256

	include asm.inc

	.text
__text:

	entry boot
	import _initialise
	import _mainloop

;----------------------------------------------------------------------
; 256 interrupt vectors at 0:0
; all handlers currently in _CS segment

	macro ivec
	.word ?1, _CS
	mend

	ivec divide_error      ; 00 0000
	ivec single_step       ; 01 0004
	ivec nmi               ; 02 0008
	ivec breakpoint        ; 03 000C
	ivec overflow          ; 04 0010
	ivec bounds_check      ; 05 0014
	ivec invalid_opcode    ; 06 0018
	ivec floating_escape   ; 07 001C
	ivec timer_0_int       ; 08 0020
	ivec reserved          ; 09 0024
	ivec dma_0_int         ; 0A 0028
	ivec dma_1_int         ; 0B 002C
	ivec int_0             ; 0C 0030
	ivec int_1             ; 0D 0034
	ivec int_2             ; 0E 0038
	ivec int_3             ; 0F 003C
	ivec reserved          ; 10 0040 (Numerics)
	ivec reserved          ; 11 0044
	ivec timer_1_int       ; 12 0048
	ivec timer_2_int       ; 13 004C
	ivec reserved          ; 14 0050
	ivec reserved          ; 15 0054
	ivec reserved          ; 16 0058
	ivec reserved          ; 17 005C
	ivec reserved          ; 18 0060
	ivec reserved          ; 19 0064
	ivec reserved          ; 1A 0068
	ivec reserved          ; 1B 006C
	ivec reserved          ; 1C 0070
	ivec reserved          ; 1D 0074
	ivec reserved          ; 1E 0078
	ivec reserved          ; 1F 007C last reserved

	ivec unknown           ; 20 0080 available for user
	ivec unknown           ; 21 0084
	ivec new_task          ; 22 0088
	ivec exit              ; 23 008C
	ivec yield             ; 24 0090
	ivec unknown 
	ivec unknown 
	ivec unknown 
	ivec unknown 
	ivec unknown 
	ivec unknown 
	ivec unknown 
	ivec unknown 
	ivec unknown 
	ivec unknown 
	ivec unknown           ; 2F 00BC

	macro eight_unknown_ivecs
	ivec unknown
	ivec unknown
	ivec unknown
	ivec unknown
	ivec unknown
	ivec unknown
	ivec unknown
	ivec unknown
	mend

	eight_unknown_ivecs    ; 30 00C0
	eight_unknown_ivecs    ; 38 00E0
	eight_unknown_ivecs    ; 40 0100
	eight_unknown_ivecs    ; 48 0120
	eight_unknown_ivecs    ; 50 0140
	eight_unknown_ivecs    ; 58 0160
	eight_unknown_ivecs    ; 60 0180
	eight_unknown_ivecs    ; 68 01A0
	eight_unknown_ivecs    ; 70 01C0
	eight_unknown_ivecs    ; 78 01E0
	eight_unknown_ivecs    ; 80 0200
	eight_unknown_ivecs    ; 88 0220
	eight_unknown_ivecs    ; 90 0240
	eight_unknown_ivecs    ; 98 0260
	eight_unknown_ivecs    ; A0 0280
	eight_unknown_ivecs    ; A8 02A0
	eight_unknown_ivecs    ; B0 02C0
	eight_unknown_ivecs    ; B8 02E0
	eight_unknown_ivecs    ; C0 0300
	eight_unknown_ivecs    ; C8 0320
	eight_unknown_ivecs    ; D0 0340
	eight_unknown_ivecs    ; D8 0360
	eight_unknown_ivecs    ; E0 0380
	eight_unknown_ivecs    ; E8 03A0
	eight_unknown_ivecs    ; F0 03C0
	eight_unknown_ivecs    ; F8 03E0

	;-------------------------------------------------
	; reset fixes UCS to 256 kB, C0000 ... FFFFF. R=101: 1 waitstate, no RDY
	; and jumps to C000:0400 (boot)
	; ss is zero, sp undefined

boot:                  ; org C000:0400 and later 0:0400
	cli
	cld

	;-------------------------------------------------
	; memory and io chipselects

	; PCS0 at 0000, R=111: 3 waitstates, no RDY, for PCS0-3

	mov ax, # 0x003F
	mov dx, # PACS
	out dx, PCB_AX

	; MCS0 at 00000, R=001: 1 waitstate, external RDY, for MCS0-3

	mov ax, # 0x01F9
	mov dx, # MMCS
	out dx, PCB_AX

	; MCS0-3: four times 128 kB = 512 kB
	; EX,MS=10: all PCS0-6 in i/o space
	; R=010: 2 waitstates, external RDY for PCS4-6

	mov ax, # 0xC0BA
	mov dx, # MPCS
	out dx, PCB_AX

	; MCS0 is at zero, so LCS can not be used. LMCS is not programmed

	;-------------------------------------------------
	; relocate from  C000:0 to ram,
	; text              0:0 64 kB and
	; data/bss/stack 1000:0 64 kB

	mov ax, # _DS      ; final SS:SP
	mov ss, ax
	mov sp, # 0        ; wraps to ss:0xFFFE on first push

	mov ax, # ROM
	mov ds, ax
	xor si, si         ; src 0xC000:0

	mov ax, # _CS
	mov es, ax
	xor di, di         ; dst 0:0

	mov cx, # __etext
	rep
	movsb              ; copy text

	xor si, si         ; rewind
	xor di, di

	mov cx, # __etext
	rep
	cmpsb              ; verify text (Vcc coming up?)
	jnz boot           ; and try again while bad copy

	; F029 + C000 + 1000 (if 29 paragraphs of text)

	mov ax, # __segoff ; includes -_DS
	add ax, # ROM + _DS
	mov ds, ax
	xor si, si         ; src (0xC000 + segoff):0

	mov ax, # _DS
	mov es, ax
	xor di, di         ; dst 0x1000:0

	mov cx, # __edata
	rep
	movsb              ; copy data

	mov cx, di
	neg cx             ; to end of segment
	xor ax, ax
	rep
	stosb              ; zero bss and stack

	jmpf .2, _CS       ; final CS
.2:
	mov ax, # _DS
	mov ds, ax         ; final DS
	mov es, ax         ; final ES

	;-------------------------------------------------
	; interrupt controller

	mov ax, # 0xFFFF   ; opened by each MSK=0 below
	mov dx, # IMASK
	out dx, PCB_AX

	mov ax, # 7        ; no levels masked
	mov dx, # IPRIO
	out dx, PCB_AX

	mov cx, # 256      ; flush pending interrupts
	mov ax, # EOI_ANY
	mov dx, # EOI
.3:
	out dx, PCB_AX
	loop .3

	;-------------------------------------------------
	; external interrupts

	mov ax, # 0x01    ; edge-trigger MSK=0 PR=1
	mov dx, # I0CON
	out dx, PCB_AX

	mov ax, # 0x10    ; level-trigger MSK=0 PR=0
	mov dx, # I1CON
	out dx, PCB_AX

	mov ax, # 0x12    ; level-trigger MSK=0 PR=2
	mov dx, # I2CON
	out dx, PCB_AX

	mov ax, # 0x13    ; level-trigger MSK=0 PR=3
	mov dx, # I3CON
	out dx, PCB_AX

	;-------------------------------------------------
	; Timer 2
	; periodic interval timer, 256 Hz, lowest priority

	mov ax, # 0x6001  ; /INH INT CONT
	mov dx, # T2MODE
	out dx, PCB_AX

	mov ax, # TMRCLK / HZ    ; approx 256.016 Hz, close enough
	mov dx, # T2CMP
	out dx, PCB_AX

	mov ax, # 0
	mov dx, # T2CNT
	out dx, PCB_AX

	mov ax, # 0xE001  ; EN /INH INT CONT
	mov dx, # T2MODE
	out dx, PCB_AX

	mov ax, # 6       ; MSK=0 PR=6
	mov dx, # TICON
	out dx, PCB_AX

	;-------------------------------------------------

	sti
	call _initialise
main:
	mov _fido, # HZ * 10    ; 10 seconds

	hlt
	call _mainloop
	jmp main

;----------------------------------------------------------------------

floating_escape:
breakpoint:
single_step:             ; XXX not worth trying to clear TF from stack
bounds_check:
overflow:
divide_error:
invalid_opcode:
reserved:

reboot:
	jmpf boot, ROM

;----------------------------------------------------------------------

	macro interrupt_prolog
	pusha
	push ds
	push es                  ; must match IREGS definition
	mov ax, # _DS
	mov ds, ax
	mov es, ax
	cld
	mend

	macro interrupt_epilog   ; optional eoi_cmd
	pop es
	pop ds
	if ?1                    ; eoi?
	cli
	mov ax, # ?1             ; EOI_ANY or EOI_xxx
	mov dx, # EOI
	out dx, PCB_AX
	endif
	popa
	iret
	mend

	macro interrupt          ; handler and optional eoi_cmd
	sti
	interrupt_prolog
	import ?1
	call ?1
	interrupt_epilog ?2
	mend

;----------------------------------------------------------------------
; unknown software interrupt

unknown:
	sti
	push bp
	mov bp, sp       ; bp ip cs flags
	or [bp+6], # 1   ; set carry
	mov ax, # -1     ; error code -1
	pop bp
	iret

;----------------------------------------------------------------------
; software interrupt 0x24
;
; no args, all regs kept

yield:                      ; interrupts left disabled throughout
	interrupt_prolog        ; IREGS to old stack

ret_next:
	mov bx, task
	mov [bx+tsk_ss], ss
	mov [bx+tsk_sp], sp

	mov bx, [bx+tsk_next]
	mov task, bx
	mov ss, [bx+tsk_ss]
	mov sp, [bx+tsk_sp]

	interrupt_epilog 0      ; IREGS from new stack

;----------------------------------------------------------------------
; software interrupt 0x23
;
; no args, no return

exit:                       ; interrupts left disabled throughout
	mov ax, # _DS
	mov ds, ax

	mov si, task
	mov ax, si
.8:
	mov bx, ax
	mov ax, [bx+tsk_next]
	cmp ax, si              ; bx is previous of si ?
	jne .8

	mov ax, [si+tsk_next]
	mov [bx+tsk_next], ax   ; removed from chain

	mov ax, task_free
	mov [si+tsk_next], ax
	mov task_free, si

	mov task, bx            ; switch to the previous
	mov ss, [bx+tsk_ss]
	mov sp, [bx+tsk_sp]

	interrupt_epilog 0      ; IREGS from new stack

;----------------------------------------------------------------------
; software interrupt 0x22
;
; arguments in stack:
;  ip_entry, cs, sp_empty, ss, ds_es
;
; returns:
;  dx:ax task structure
;
; bx, cx trashed

new_task:
	sti                     ; benign stuff first
	push bp
	mov bp, sp              ; bp ip cs flags ip cs sp ss dses
	push si                 ; 0  2  4  6     8  10 12 14 16
	push ds

	; prepare initial registers

	mov ax, [bp+8]          ; ip_entry
	mov cx, [bp+10]         ; cs
	lds bx, [bp+12]         ; sp_empty ss
	mov dx, [bp+16]         ; ds_es

	sub bx, # IREGS_SZ      ; regs on stacktop

	mov [bx+IREGS_DS], dx   ; DS=ES
	mov [bx+IREGS_ES], dx

	mov [bx+IREGS_IP], ax
	mov [bx+IREGS_CS], cx
	mov [bx+IREGS_FL], # 0x200 ; IF. interrupts will be enabled

	xor ax, ax              ; lots of zeroes
	mov [bx+IREGS_AX], ax
	mov [bx+IREGS_BX], ax
	mov [bx+IREGS_CX], ax
	mov [bx+IREGS_DX], ax
	mov [bx+IREGS_SI], ax
	mov [bx+IREGS_DI], ax
	mov [bx+IREGS_BP], ax

	mov cx, ds              ; stacktop in cx:dx
	mov dx, bx

	; link to task chain

	mov ax, # _DS
	mov ds, ax

	cli

	mov bx, task_free
	or bx, bx
	jz .5                   ; must allocate new task structure
	mov ax, [bx+tsk_next]
	mov task_free, ax
	jmp .6
.5:
	mov bx, _brk
	mov ax, bx
	add ax, # tsk_size
	mov _brk, ax
.6:
	mov [bx+tsk_zzz], # 0

	mov [bx+tsk_ss], cx
	mov [bx+tsk_sp], dx

	mov si, task            ; link new task after current one
	mov ax, [si+tsk_next]
	mov [bx+tsk_next], ax
	mov [si+tsk_next], bx

	clc                     ; success
	mov dx, # _DS
	mov ax, bx

	pop ds
	pop si
	pop bp
	iret

;----------------------------------------------------------------------

nmi:
	interrupt _nmi, 0

dma_0_int:
	interrupt _dma_0_int, EOI_DMA0

dma_1_int:
	interrupt _dma_1_int, EOI_DMA1

int_1:
	interrupt _int_1, EOI_INT1

int_3:
	interrupt _int_3, EOI_INT3

timer_0_int:
	interrupt _timer_0_int, EOI_TMR

timer_1_int:
	interrupt _timer_1_int, EOI_TMR

;----------------------------------------------------------------------
; 5 uarts in the asic

	macro poll_uart    ; uart_index source_bit data_reg

	import _uart_?1_int

	test al, # ?2      ; activity on this uart ?
	jz not_0_?1

	in al, ?3 + 2      ; uart status, 2 tint, 1 rint
	mov ah, al         ; pass status in msbyte
	test al, # 1       ; rint bit
	jz not_rint_?1

	in al, ?3 + 0      ; pass uart data in lsbyte
not_rint_?1:
	call _uart_?1_int

	mov ax, bp         ; XXX
not_0_?1:
	mend

int_0:
	sti
	interrupt_prolog

	in al, 0x01               ; int_0 sources
again0:
	mov bp, ax                ; XXX

	poll_uart 0, 0x01, 0x04
	poll_uart 1, 0x02, 0x08
	poll_uart 2, 0x04, 0x0C
	poll_uart 3, 0x08, 0x10
	poll_uart 4, 0x10, 0x14

	in al, 0x01               ; int_0 sources
	and al, # 0x1F
	jnz again0

	interrupt_epilog EOI_INT0

;----------------------------------------------------------------------
; 4 other serial devices in the asic
; input-only devices ?
;
; keyboard
; keypad
;
; maybe:
; barcode reader ?
; vertical blanking ?
; modem inputs ?

int_2:
	sti
	interrupt_prolog

	in al, 0x03           ; int_2 sources
again2:
	mov bp, ax            ; XXX

	test al, # 1
	jz not_2_0

	in al, 0x21           ; ?
	in al, 0x20           ; ? crtc related?

	mov ax, bp
not_2_0:
	test al, # 2
	jz not_2_1

	in al, 0x25           ; guess:
	mov ah, al            ; these 8 input bits have changed, to
	in al, 0x24           ; these states

	mov ax, bp
not_2_1:
	test al, # 4
	jz not_2_2

	in al, 0x29           ; ?
	test al, # 0x08
	jz dunno
	in al, 0x29           ; ?
dunno:

	mov ax, bp
not_2_2:
	test al, # 8
	jz not_2_3

	in al, 0x18           ; ?

	mov ax, bp
not_2_3:

	in al, 0x3            ; int_2 sources
	and al, # 0x0F
	jnz again2

	interrupt_epilog EOI_INT2

;----------------------------------------------------------------------
; 256 Hz periodic, lowest priority

timer_2_int:
	interrupt_prolog

	mov ax, # SCREEN
	mov ds, ax
	incb [80]
	jnz foo1
	incb [82]
	jnz foo1
	incb [84]
foo1:
	mov ax, # _DS
	mov ds, ax

	mov al, # 1           ; XXX watchdog in ASIC
	out 0x2A, al

	dec _fido
	jz reboot_

	mov bx, # _timex      ; bump timex by 1/256 seconds
	add [bx+0], # 65536 / HZ
	adc [bx+2], # 0
	sti

	import _tick
	call _tick
	interrupt_epilog EOI_TMR

reboot_:
	jmpf boot, ROM

;----------------------------------------------------------------------
	export _disable, _enable
	export _spl_up, _splx
	export _outb, _outw
	export _inb, _inw

_disable:             ; word disable();
	pushf
	cli
	pop ax
	ret

_restore:             ; void restore(word oflags);
	and ax, # 0x200   ; ARG1. FLAGS.IF ?
	jz .4
_enable:              ; void enable();
	sti
.4:
	ret

_splx:                ; void splx(opri);
	mov dx, # IPRIO
	out dx, PCB_AX    ; ARG1
	ret

_spl_up:              ; word spl_up(npri);
	mov bx, ax        ; ARG1
	mov dx, # IPRIO
	in ax, dx
	cmp bx, ax
	jae .1
	xchg ax, bx
	out dx, PCB_AX
	xchg ax, bx
.1:
	ret

_inb:                 ; word inb(port);
	mov dx, ax        ; ARG1
	in al, dx
	xor ah, ah        ; zeroextend
	ret

_inw:                 ; word inw(port);
	mov dx, ax        ; ARG1
	in ax, dx
	ret

_outb:                ; void outb(port, byte);
	mov dx, ax        ; ARG1
	ARG2 al
	out dx, al
	ret

_outw:                ; void outw(port, word);
	mov dx, ax        ; ARG1
	ARG2 ax
	out dx, PCB_AX
	ret

;----------------------------------------------------------------------
	.data
__data:

	.long __text
	.long __data
	.long __bss
_brk:
	.long __end

	.asciz VERSION

task:
	.word task_0

task_free:
	.word 0

tsk_next equ 0
tsk_sp   equ 2
tsk_ss   equ 4
tsk_zzz  equ 6      ; unused
tsk_size equ 8

task_0:
	.word task_0    ; next.   now only task_0 exists
	.word 0, _DS    ; sp, ss. filled when switching out
	.word 0         ; unused

_cursor:
	.word 0
	.word SCREEN

;----------------------------------------------------------------------
	.bss
__bss:

	export _timex, _timef, _time  ; seconds, in different formats
_timex:
	.space 1
_timef:
	.space 1                      ; 8.8 or 16.8
_time:
	.space 2                      ; 16.0

zz	lcomm 7

_fido:
	.space 2

	.comm _zz, 7
