
#include "head.h"
#include "pcb.h"
#include "io.h"
#include "crtc.h"

static void
zappa(pp, b)
	byte **pp, b;
{
	register byte *p = *pp;

	p[0] = hexdig[b / 16];
	p[1] = hexdig[b % 16];
	p[2] = ' ';

	*pp += 3;
}

/* serport 2/3 */

void
putc(b)
{
	while (inb(0x0A) & 0x8)
		;
	outb(0x08, b);
}

void
putc_hex(b)
{
	putc(hexdig[b / 16]);
	putc(hexdig[b % 16]);
}

void
regshow()
{
	byte buf[40], *p = buf;

	zappa(&p, inb(0x01));

	zappa(&p, inb(0x06));
#if 0
	zappa(&p, inb(0x0A));  /* status ? uart at d25 pins 2/3 */
#endif
	zappa(&p, inb(0x0E));

	zappa(&p, inb(0x12));  /* uart at d25 pins ?/17 */

	zappa(&p, inb(0x16));

	zappa(&p, inb(0x18));
	zappa(&p, inb(0x21));
	zappa(&p, inb(0x20));
	zappa(&p, inb(0x24)); /* low byte */
	zappa(&p, inb(0x25)); /* high byte */
	zappa(&p, inb(0x29));

	update_row(3, buf);
}

void
initialise()
{
	byte buf[40], *p = buf;

	/* 40 character columns, 16 rows, 240 vertical pixels */

	clear_screen();

	outw_crtc(VPIXELS, 240 - 1); /* Nd - 1 */
	outw_crtc(CURSOR_ADDR, 0);
	outw_crtc(START_ADDR, 0);

	outb_crtc(HCHR_TOT, 56 - 1); /* Nht - 1 */
	outb_crtc(HCHR_DISPLAY, 40); /* Nhd */
	outb_crtc(HCHR_VIRTUAL, 40); /* Nir */

	outb_crtc(START_RASTER, 0);             /* Nsr */
	outb_crtc(CHR_END, 14);                 /* Nr  */
	outb_crtc(CURSOR_END, 7);               /* Nce */
	outb_crtc(CURSOR_START, 6 | CURSOR_ON); /* Ncs + blink mode */

	outb_crtc(MODE, MODE_AT | MODE_ON);

	update_row(0, "Zanzibar                                ");

	/* serial ports */

	/*
	 *  0 r and t data
	 *  1 config, 0x20 and 3 other bitfields. 0x06 = 19200bd
	 *  2 status
	 *  3 intena bitfield
	 */

	outb(0x09, 0x06); /* 0x20? 0x? 0x? 300bd (defaulted) to 19200 */
	outb(0x0B, 0x01); /* rxintena=1 txintena=0 ? */

	outb(0x11, 0x06); /* 0x20? 0x? 0x? 300bd (defaulted) to 19200 */
	outb(0x13, 0x01); /* rxintena=1 txintena=0 ? */

	/* regshow title */

	zappa(&p, 0x01);

	zappa(&p, 0x06);
#if 0
	zappa(&p, 0x0A);  /* status ? uart at d25 pins 2/3 */
#endif
	zappa(&p, 0x0E);

	zappa(&p, 0x12);  /* uart at d25 pins ?/17 */

	zappa(&p, 0x16);

	zappa(&p, 0x18);
	zappa(&p, 0x21);
	zappa(&p, 0x20);
	zappa(&p, 0x24); /* low byte */
	zappa(&p, 0x25); /* high byte */
	zappa(&p, 0x29);

	update_row(2, buf);
}

void
mainloop()
{
	byte x, c;
	static word otim;

	regshow();

#if 0
	/*
	 * 0x1 rxint ?
	 * 0x2 txint ? error int?
	 * 0x4 RXBF
	 * 0x8 TXBF
	 * 0x10 parity maybe, or other error (0x70 seem to be errors)
	 * 0x20 overrun ?
	 * 0x40 break ? framing ?
	 * 0x80
	 */

	x = inb(0x0A);

	if (x & ~0xC) {
		putc('E');
		putc('=');
		putc_hex(x);
		putc(' ');
	}
	if (x & 0x4) {
		putc(inb(0x08));
	}

	if (otim != time) {
		otim = time;

		outb(?, otim & ?); /* toggle something */
	}
#endif
}

void
nmi(IREGS)
{
}

void
tick(IREGS)        /* 256 Hz periodic from timer_2_int */
{
}

/*
 *  int_0 is in boot.c and
 *  directly calls each uart_N_int() in uart.c
 *
 *  int_2 will be in boot.c too XXX
 */

void
int_2(IREGS)
{
	byte x;

	while (x = inb(0x03)) {
		if (x & 0x1) {
			inb(0x21);
			inb(0x20);
		}
		if (x & 0x2) {
			inb(0x25); /* bit changes */
			inb(0x24); /* bit values */
		}
		if (x & 0x4) {
			if (inb(0x29) & 0x8)
				inb(0x29);
		}
		if (x & 0x8) {
			inb(0x18);
		}
	}
}

void
int_1(IREGS)
{
}

void
int_3(IREGS)
{
}

void
timer_0_int(IREGS)
{
}

void
timer_1_int(IREGS)
{
}

void
dma_0_int(IREGS)
{
}

void
dma_1_int(IREGS)
{
}

